# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import logging
import os
import shutil
import threading
import time

import watchdog.events

import octoprint.filemanager
import octoprint.filemanager.util
import octoprint.util


def fix_scandir():
    try:
        from os import scandir

        # nothing to fix, natively available, return
        return

    except ImportError:
        # not natively available, use backport
        import watchdog.utils.dirsnapshot
        from scandir import scandir

        OriginalDirectorySnapshot = watchdog.utils.dirsnapshot.DirectorySnapshot

        class FixedDirectorySnapshot(OriginalDirectorySnapshot):
            def __init__(self, listdir=scandir, *args, **kwargs):
                OriginalDirectorySnapshot.__init__(self, listdir=listdir, *args, **kwargs)

        watchdog.utils.dirsnapshot.DirectorySnapshot = FixedDirectorySnapshot

        import watchdog.observers.polling

        OriginalPollingEmitter = watchdog.observers.polling.PollingEmitter

        class FixedPollingEmitter(OriginalPollingEmitter):
            def __init__(self, listdir=scandir, *args, **kwargs):
                OriginalPollingEmitter.__init__(self, listdir=listdir, *args, **kwargs)

        watchdog.observers.polling.PollingEmitter = FixedPollingEmitter


class GcodeWatchdogHandler(watchdog.events.PatternMatchingEventHandler):

    """
    Takes care of automatically "uploading" files that get added to the watched folder.
    """

    def __init__(self, file_manager, printer):
        watchdog.events.PatternMatchingEventHandler.__init__(
            self,
            patterns=list(
                map(lambda x: "*.%s" % x, octoprint.filemanager.get_all_extensions())
            ),
        )

        self._logger = logging.getLogger(__name__)

        self._file_manager = file_manager
        self._printer = printer

        self._watched_folder = None

    def initial_scan(self, folder):
        try:
            from os import scandir
        except ImportError:
            from scandir import scandir

        def _recursive_scandir(path):
            """Recursively yield DirEntry objects for given directory."""
            for entry in scandir(path):
                if entry.is_dir(follow_symlinks=False):
                    for entry in _recursive_scandir(entry.path):
                        yield entry
                else:
                    yield entry

        def run_scan():
            self._logger.info("Running initial scan on watched folder...")
            self._watched_folder = folder

            for entry in _recursive_scandir(folder):
                path = entry.path

                if not self._valid_path(path):
                    continue

                self._logger.info("Found {}, trying to add it".format(path))
                self._upload(path)
            self._logger.info("... initial scan done.")

        thread = threading.Thread(target=run_scan)
        thread.daemon = True
        thread.start()

    def on_created(self, event):
        self._start_check(event.src_path)

    def on_moved(self, event):
        # for move/rename we are only interested in the dest_path
        self._start_check(event.dest_path)

    def _start_check(self, path):
        if not self._valid_path(path):
            return

        thread = threading.Thread(target=self._repeatedly_check, args=(path,))
        thread.daemon = True
        thread.start()

    def _upload(self, path):
        # noinspection PyBroadException
        try:
            file_wrapper = octoprint.filemanager.util.DiskFileWrapper(
                os.path.basename(path), path
            )
            relative_path = os.path.relpath(path, self._watched_folder)

            # determine future filename of file to be uploaded, abort if it can't be uploaded
            try:
                futurePath, futureFilename = self._file_manager.sanitize(
                    octoprint.filemanager.FileDestinations.LOCAL, relative_path
                )
            except Exception:
                self._logger.exception("Could not wrap %s", path)
                futurePath = None
                futureFilename = None

            if futureFilename is None or (
                len(self._file_manager.registered_slicers) == 0
                and not octoprint.filemanager.valid_file_type(futureFilename)
            ):
                return

            # prohibit overwriting currently selected file while it's being printed
            futureFullPath = self._file_manager.join_path(
                octoprint.filemanager.FileDestinations.LOCAL, futurePath, futureFilename
            )
            futureFullPathInStorage = self._file_manager.path_in_storage(
                octoprint.filemanager.FileDestinations.LOCAL, futureFullPath
            )

            if not self._printer.can_modify_file(futureFullPathInStorage, False):
                return

            reselect = self._printer.is_current_file(futureFullPathInStorage, False)

            added_file = self._file_manager.add_file(
                octoprint.filemanager.FileDestinations.LOCAL,
                relative_path,
                file_wrapper,
                allow_overwrite=True,
            )
            if os.path.exists(path):
                try:
                    os.remove(path)
                except Exception:
                    pass

            if reselect:
                self._printer.select_file(
                    self._file_manager.path_on_disk(
                        octoprint.filemanager.FileDestinations.LOCAL, added_file
                    ),
                    False,
                )
        except Exception:
            self._logger.exception(
                "There was an error while processing the file {} in the watched folder".format(
                    path
                )
            )
        finally:
            if os.path.exists(path):
                # file is still there - that should only happen if something went wrong, so mark it as failed
                # noinspection PyBroadException
                try:
                    shutil.move(path, "{}.failed".format(path))
                except Exception:
                    # something went really wrong here.... but we can't do anything about it, so just log it
                    self._logger.exception(
                        "There was an error while trying to mark {} as failed in the watched folder".format(
                            path
                        )
                    )

    def _repeatedly_check(self, path, interval=1, stable=5):
        try:
            last_size = os.stat(path).st_size
        except Exception:
            return

        countdown = stable

        while True:
            try:
                new_size = os.stat(path).st_size
            except Exception:
                return

            if new_size == last_size:
                self._logger.debug(
                    "File at {} is no longer growing, counting down: {}".format(
                        path, countdown
                    )
                )
                countdown -= 1
                if countdown <= 0:
                    break
            else:
                self._logger.debug(
                    "File at {} is still growing (last: {}, new: {}), waiting...".format(
                        path, last_size, new_size
                    )
                )
                countdown = stable

            last_size = new_size
            time.sleep(interval)

        self._logger.debug("File at {} is stable, moving it".format(path))
        self._upload(path)

    def _valid_path(self, path):
        _, ext = os.path.splitext(path)
        return (
            octoprint.filemanager.valid_file_type(path)
            and not octoprint.util.is_hidden_path(path)
            and ext != "failed"
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Marc Hannappel <salandora@gmail.com>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2017 The OctoPrint Project - Released under terms of the AGPLv3 License"

# Wrapper to the new access.users location

import warnings

from octoprint.access.users import *  # noqa: F401, F403 ## possibly used by other modules
from octoprint.access.users import User, deprecated

warnings.warn(
    "octoprint.users is deprecated, use octoprint.access.users instead",
    DeprecationWarning,
    stacklevel=2,
)

AccessUser = User


class User(AccessUser):
    @deprecated(
        "octoprint.users.User is deprecated, please use octoprint.access.users.User instead"
    )
    def __init__(self, username, passwordHash, active, roles, apikey=None, settings=None):
        from octoprint.server import groupManager

        if "admin" in roles:
            groups = [groupManager.admin_group]
        elif "user" in roles:
            groups = [groupManager.user_group]
        else:
            groups = [groupManager.guest_group]

        AccessUser(
            username=username,
            passwordHash=passwordHash,
            active=active,
            permissions=None,
            groups=groups,
            apikey=apikey,
            settings=settings,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

if __name__ == "__main__":
    import octoprint

    octoprint.main()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
In this module the slicing support of OctoPrint is encapsulated.

.. autoclass:: SlicingProfile
   :members:

.. autoclass:: TemporaryProfile
   :members:

.. autoclass:: SlicingManager
   :members:
"""

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


import logging
import os

import octoprint.events
import octoprint.plugin
import octoprint.util
from octoprint.settings import settings

from .exceptions import (
    CouldNotDeleteProfile,
    ProfileAlreadyExists,
    ProfileException,
    SlicerNotConfigured,
    SlicingCancelled,
    UnknownProfile,
    UnknownSlicer,
)


class SlicingProfile(object):
    """
    A wrapper for slicing profiles, both meta data and actual profile data.

    Arguments:
        slicer (str): Identifier of the slicer this profile belongs to.
        name (str): Identifier of this slicing profile.
        data (object): Profile data, actual structure depends on individual slicer implementation.
        display_name (str): Displayable name for this slicing profile.
        description (str): Description of this slicing profile.
        default (bool): Whether this is the default slicing profile for the slicer.
    """

    def __init__(
        self, slicer, name, data, display_name=None, description=None, default=False
    ):
        self.slicer = slicer
        self.name = name
        self.data = data
        self.display_name = display_name
        self.description = description
        self.default = default


class TemporaryProfile(object):
    """
    A wrapper for a temporary slicing profile to be used for a slicing job, based on a :class:`SlicingProfile` with
    optional ``overrides`` applied through the supplied ``save_profile`` method.

    Usage example:

    .. code-block:: python

       temporary = TemporaryProfile(my_slicer.save_slicer_profile, my_default_profile,
                                    overrides=my_overrides)
       with (temporary) as profile_path:
           my_slicer.do_slice(..., profile_path=profile_path, ...)

    Arguments:
        save_profile (callable): Method to use for saving the temporary profile, also responsible for applying the
            supplied ``overrides``. This will be called according to the method signature of
            :meth:`~octoprint.plugin.SlicerPlugin.save_slicer_profile`.
        profile (SlicingProfile): The profile from which to derive the temporary profile.
        overrides (dict): Optional overrides to apply to the ``profile`` for creation of the temporary profile.
    """

    def __init__(self, save_profile, profile, overrides=None):
        self.save_profile = save_profile
        self.profile = profile
        self.overrides = overrides

    def __enter__(self):
        import tempfile

        temp_profile = tempfile.NamedTemporaryFile(
            prefix="slicing-profile-temp-", suffix=".profile", delete=False
        )
        temp_profile.close()

        self.temp_path = temp_profile.name
        self.save_profile(self.temp_path, self.profile, overrides=self.overrides)
        return self.temp_path

    def __exit__(self, type, value, traceback):
        import os

        try:
            os.remove(self.temp_path)
        except Exception:
            pass


class SlicingManager(object):
    """
    The :class:`SlicingManager` is responsible for managing available slicers and slicing profiles.

    Arguments:
        profile_path (str): Absolute path to the base folder where all slicing profiles are stored.
        printer_profile_manager (~octoprint.printer.profile.PrinterProfileManager): :class:`~octoprint.printer.profile.PrinterProfileManager`
           instance to use for accessing available printer profiles, most importantly the currently selected one.
    """

    def __init__(self, profile_path, printer_profile_manager):
        self._logger = logging.getLogger(__name__)

        self._profile_path = profile_path
        self._printer_profile_manager = printer_profile_manager

        self._slicers = {}
        self._slicer_names = {}

    def initialize(self):
        """
        Initializes the slicing manager by loading and initializing all available
        :class:`~octoprint.plugin.SlicerPlugin` implementations.
        """
        self.reload_slicers()

    def reload_slicers(self):
        """
        Retrieves all registered :class:`~octoprint.plugin.SlicerPlugin` implementations and registers them as
        available slicers.
        """
        plugins = octoprint.plugin.plugin_manager().get_implementations(
            octoprint.plugin.SlicerPlugin
        )
        slicers = {}
        for plugin in plugins:
            try:
                slicers[plugin.get_slicer_properties()["type"]] = plugin
            except Exception:
                self._logger.exception(
                    "Error while getting properties from slicer {}, ignoring it".format(
                        plugin._identifier
                    ),
                    extra={"plugin": plugin._identifier},
                )
                continue
        self._slicers = slicers

    @property
    def slicing_enabled(self):
        """
        Returns:
            (boolean) True if there is at least one configured slicer available, False otherwise.
        """
        return len(self.configured_slicers) > 0

    @property
    def registered_slicers(self):
        """
        Returns:
            (list of str) Identifiers of all available slicers.
        """
        return list(self._slicers.keys())

    @property
    def configured_slicers(self):
        """
        Returns:
            (list of str) Identifiers of all available configured slicers.
        """
        return list(
            map(
                lambda slicer: slicer.get_slicer_properties()["type"],
                filter(
                    lambda slicer: slicer.is_slicer_configured(), self._slicers.values()
                ),
            )
        )

    @property
    def default_slicer(self):
        """
        Retrieves the default slicer.

        Returns:
            (str) The identifier of the default slicer or ``None`` if the default slicer is not registered in the
                system.
        """
        slicer_name = settings().get(["slicing", "defaultSlicer"])
        if slicer_name in self.registered_slicers:
            return slicer_name
        else:
            return None

    def get_slicer(self, slicer, require_configured=True):
        """
        Retrieves the slicer named ``slicer``. If ``require_configured`` is set to True (the default) an exception
        will be raised if the slicer is not yet configured.

        Arguments:
            slicer (str): Identifier of the slicer to return
            require_configured (boolean): Whether to raise an exception if the slicer has not been configured yet (True,
                the default), or also return an unconfigured slicer (False).

        Raises:
            ~octoprint.slicing.exceptions.UnknownSlicer: The ``slicer`` is unknown.
            ~octoprint.slicing.exceptions.SlicerNotConfigured: The ``slicer`` is not yet configured and ``require_configured`` was set to True.
        """

        if slicer not in self._slicers:
            raise UnknownSlicer(slicer)

        if require_configured and not self._slicers[slicer].is_slicer_configured():
            raise SlicerNotConfigured(slicer)

        return self._slicers[slicer]

    def slice(
        self,
        slicer_name,
        source_path,
        dest_path,
        profile_name,
        callback,
        callback_args=None,
        callback_kwargs=None,
        overrides=None,
        on_progress=None,
        on_progress_args=None,
        on_progress_kwargs=None,
        printer_profile_id=None,
        position=None,
    ):
        """
        Slices ``source_path`` to ``dest_path`` using slicer ``slicer_name`` and slicing profile ``profile_name``.
        Since slicing happens asynchronously, ``callback`` will be called when slicing has finished (either successfully
        or not), with ``callback_args`` and ``callback_kwargs`` supplied.

        If ``callback_args`` is left out, an empty argument list will be assumed for the callback. If ``callback_kwargs``
        is left out, likewise an empty keyword argument list will be assumed for the callback. Note that in any case
        the callback *must* support being called with the following optional keyword arguments:

        _analysis
            If the slicer returned analysis data of the created machine code as part of its slicing result, this keyword
            argument will contain that data.
        _error
            If there was an error while slicing this keyword argument will contain the error message as returned from
            the slicer.
        _cancelled
            If the slicing job was cancelled this keyword argument will be set to True.

        Additionally callees may specify ``overrides`` for the specified slicing profile, e.g. a different extrusion
        temperature than defined in the profile or a different layer height.

        With ``on_progress``, ``on_progress_args`` and ``on_progress_kwargs``, callees may specify a callback plus
        arguments and keyword arguments to call upon progress reports from the slicing job. The progress callback will
        be called with a keyword argument ``_progress`` containing the current slicing progress as a value between 0
        and 1 plus all additionally specified args and kwargs.

        If a different printer profile than the currently selected one is to be used for slicing, its id can be provided
        via the keyword argument ``printer_profile_id``.

        If the ``source_path`` is to be a sliced at a different position than the print bed center, this ``position`` can
        be supplied as a dictionary defining the ``x`` and ``y`` coordinate in print bed coordinates of the model's center.

        Arguments:
            slicer_name (str): The identifier of the slicer to use for slicing.
            source_path (str): The absolute path to the source file to slice.
            dest_path (str): The absolute path to the destination file to slice to.
            profile_name (str): The name of the slicing profile to use.
            callback (callable): A callback to call after slicing has finished.
            callback_args (list or tuple): Arguments of the callback to call after slicing has finished. Defaults to
                an empty list.
            callback_kwargs (dict): Keyword arguments for the callback to call after slicing has finished, will be
                extended by ``_analysis``, ``_error`` or ``_cancelled`` as described above! Defaults to an empty
                dictionary.
            overrides (dict): Overrides for the printer profile to apply.
            on_progress (callable): Callback to call upon slicing progress.
            on_progress_args (list or tuple): Arguments of the progress callback. Defaults to an empty list.
            on_progress_kwargs (dict): Keyword arguments of the progress callback, will be extended by ``_progress``
                as described above! Defaults to an empty dictionary.
            printer_profile_id (str): Identifier of the printer profile for which to slice, if another than the
                one currently selected is to be used.
            position (dict): Dictionary containing the ``x`` and ``y`` coordinate in the print bed's coordinate system
                of the sliced model's center. If not provided the model will be positioned at the print bed's center.
                Example: ``dict(x=10,y=20)``.

        Raises:
            ~octoprint.slicing.exceptions.UnknownSlicer: The slicer specified via ``slicer_name`` is unknown.
            ~octoprint.slicing.exceptions.SlicerNotConfigured: The slice specified via ``slicer_name`` is not configured yet.
        """

        if callback_args is None:
            callback_args = ()
        if callback_kwargs is None:
            callback_kwargs = {}

        if slicer_name not in self.configured_slicers:
            if slicer_name not in self.registered_slicers:
                error = "No such slicer: {slicer_name}".format(**locals())
                exc = UnknownSlicer(slicer_name)
            else:
                error = "Slicer not configured: {slicer_name}".format(**locals())
                exc = SlicerNotConfigured(slicer_name)
            callback_kwargs.update({"_error": error, "_exc": exc})
            callback(*callback_args, **callback_kwargs)
            raise exc

        slicer = self.get_slicer(slicer_name)

        printer_profile = None
        if printer_profile_id is not None:
            printer_profile = self._printer_profile_manager.get(printer_profile_id)

        if printer_profile is None:
            printer_profile = self._printer_profile_manager.get_current_or_default()

        def slicer_worker(
            slicer,
            model_path,
            machinecode_path,
            profile_name,
            overrides,
            printer_profile,
            position,
            callback,
            callback_args,
            callback_kwargs,
        ):
            try:
                slicer_name = slicer.get_slicer_properties()["type"]
                with self._temporary_profile(
                    slicer_name, name=profile_name, overrides=overrides
                ) as profile_path:
                    ok, result = slicer.do_slice(
                        model_path,
                        printer_profile,
                        machinecode_path=machinecode_path,
                        profile_path=profile_path,
                        position=position,
                        on_progress=on_progress,
                        on_progress_args=on_progress_args,
                        on_progress_kwargs=on_progress_kwargs,
                    )

                if not ok:
                    callback_kwargs.update({"_error": result})
                elif (
                    result is not None
                    and isinstance(result, dict)
                    and "analysis" in result
                ):
                    callback_kwargs.update({"_analysis": result["analysis"]})
            except SlicingCancelled:
                callback_kwargs.update({"_cancelled": True})
            finally:
                callback(*callback_args, **callback_kwargs)

        import threading

        slicer_worker_thread = threading.Thread(
            target=slicer_worker,
            args=(
                slicer,
                source_path,
                dest_path,
                profile_name,
                overrides,
                printer_profile,
                position,
                callback,
                callback_args,
                callback_kwargs,
            ),
        )
        slicer_worker_thread.daemon = True
        slicer_worker_thread.start()

    def cancel_slicing(self, slicer_name, source_path, dest_path):
        """
        Cancels the slicing job on slicer ``slicer_name`` from ``source_path`` to ``dest_path``.

        Arguments:
            slicer_name (str): Identifier of the slicer on which to cancel the job.
            source_path (str): The absolute path to the source file being sliced.
            dest_path (str): The absolute path to the destination file being sliced to.

        Raises:
            ~octoprint.slicing.exceptions.UnknownSlicer: The slicer specified via ``slicer_name`` is unknown.
        """

        slicer = self.get_slicer(slicer_name)
        slicer.cancel_slicing(dest_path)

    def load_profile(self, slicer, name, require_configured=True):
        """
        Loads the slicing profile for ``slicer`` with the given profile ``name`` and returns it. If it can't be loaded
        due to an :class:`IOError` ``None`` will be returned instead.

        If ``require_configured`` is True (the default) a :class:`SlicerNotConfigured` exception will be raised
        if the indicated ``slicer`` has not yet been configured.

        Returns:
            SlicingProfile: The requested slicing profile or None if it could not be loaded.

        Raises:
            ~octoprint.slicing.exceptions.UnknownSlicer: The slicer specified via ``slicer`` is unknown.
            ~octoprint.slicing.exceptions.SlicerNotConfigured: The slicer specified via ``slicer`` has not yet been configured and
                ``require_configured`` was True.
            ~octoprint.slicing.exceptions.UnknownProfile: The profile for slicer ``slicer`` named ``name`` does not exist.
        """

        if slicer not in self.registered_slicers:
            raise UnknownSlicer(slicer)

        try:
            path = self.get_profile_path(slicer, name, must_exist=True)
        except IOError:
            return None
        return self._load_profile_from_path(
            slicer, path, require_configured=require_configured
        )

    def save_profile(
        self,
        slicer,
        name,
        profile,
        overrides=None,
        allow_overwrite=True,
        display_name=None,
        description=None,
    ):
        """
        Saves the slicer profile ``profile`` for slicer ``slicer`` under name ``name``.

        ``profile`` may be either a :class:`SlicingProfile` or a :class:`dict`.

        If it's a :class:`SlicingProfile`, its :attr:`~SlicingProfile.slicer``, :attr:`~SlicingProfile.name` and - if
        provided - :attr:`~SlicingProfile.display_name` and :attr:`~SlicingProfile.description` attributes will be
        overwritten with the supplied values.

        If it's a :class:`dict`, a new :class:`SlicingProfile` instance will be created with the supplied meta data and
        the profile data as the :attr:`~SlicingProfile.data` attribute.

        .. note::

           If the profile is the first profile to be saved for the slicer, it will automatically be marked as default.

        Arguments:
            slicer (str): Identifier of the slicer for which to save the ``profile``.
            name (str): Identifier under which to save the ``profile``.
            profile (SlicingProfile or dict): The :class:`SlicingProfile` or a :class:`dict` containing the profile
                data of the profile the save.
            overrides (dict): Overrides to apply to the ``profile`` before saving it.
            allow_overwrite (boolean): If True (default) if a profile for the same ``slicer`` of the same ``name``
                already exists, it will be overwritten. Otherwise an exception will be thrown.
            display_name (str): The name to display to the user for the profile.
            description (str): A description of the profile.

        Returns:
            SlicingProfile: The saved profile (including the applied overrides).

        Raises:
            ValueError: The supplied ``profile`` is neither a :class:`SlicingProfile` nor a :class:`dict`.
            ~octoprint.slicing.exceptions.UnknownSlicer: The slicer ``slicer`` is unknown.
            ~octoprint.slicing.exceptions.ProfileAlreadyExists: A profile with name ``name`` already exists for ``slicer`` and ``allow_overwrite`` is
                False.
        """
        if slicer not in self.registered_slicers:
            raise UnknownSlicer(slicer)

        if not isinstance(profile, SlicingProfile):
            if isinstance(profile, dict):
                profile = SlicingProfile(
                    slicer,
                    name,
                    profile,
                    display_name=display_name,
                    description=description,
                )
            else:
                raise ValueError("profile must be a SlicingProfile or a dict")
        else:
            profile.slicer = slicer
            profile.name = name
            if display_name is not None:
                profile.display_name = display_name
            if description is not None:
                profile.description = description

        first_profile = len(self.all_profiles(slicer, require_configured=False)) == 0

        path = self.get_profile_path(slicer, name)
        is_overwrite = os.path.exists(path)

        if is_overwrite and not allow_overwrite:
            raise ProfileAlreadyExists(slicer, profile.name)

        self._save_profile_to_path(
            slicer, path, profile, overrides=overrides, allow_overwrite=allow_overwrite
        )

        payload = {"slicer": slicer, "profile": name}
        event = (
            octoprint.events.Events.SLICING_PROFILE_MODIFIED
            if is_overwrite
            else octoprint.events.Events.SLICING_PROFILE_ADDED
        )
        octoprint.events.eventManager().fire(event, payload)

        if first_profile:
            # enforce the first profile we add for this slicer  is set as default
            self.set_default_profile(slicer, name)

        return profile

    def _temporary_profile(self, slicer, name=None, overrides=None):
        if slicer not in self.registered_slicers:
            raise UnknownSlicer(slicer)

        profile = self._get_default_profile(slicer)
        if name:
            try:
                profile = self.load_profile(slicer, name)
            except (UnknownProfile, IOError):
                # in that case we'll use the default profile
                pass

        return TemporaryProfile(
            self.get_slicer(slicer).save_slicer_profile, profile, overrides=overrides
        )

    def delete_profile(self, slicer, name):
        """
        Deletes the profile ``name`` for the specified ``slicer``.

        If the profile does not exist, nothing will happen.

        Arguments:
            slicer (str): Identifier of the slicer for which to delete the profile.
            name (str): Identifier of the profile to delete.

        Raises:
            ~octoprint.slicing.exceptions.UnknownSlicer: The slicer ``slicer`` is unknown.
            ~octoprint.slicing.exceptions.CouldNotDeleteProfile: There was an error while deleting the profile.
        """

        if slicer not in self.registered_slicers:
            raise UnknownSlicer(slicer)

        if not name:
            raise ValueError("name must be set")

        try:
            try:
                path = self.get_profile_path(slicer, name, must_exist=True)
            except UnknownProfile:
                return
            os.remove(path)
        except ProfileException as e:
            raise e
        except Exception as e:
            raise CouldNotDeleteProfile(slicer, name, cause=e)
        else:
            octoprint.events.eventManager().fire(
                octoprint.events.Events.SLICING_PROFILE_DELETED,
                {"slicer": slicer, "profile": name},
            )

    def set_default_profile(
        self, slicer, name, require_configured=False, require_exists=True
    ):
        """
        Sets the given profile as default profile for the slicer.

        Arguments:
            slicer (str): Identifier of the slicer for which to set the default
                profile.
            name (str): Identifier of the profile to set as default.
            require_configured (bool): Whether the slicer needs to be configured
                for the action to succeed. Defaults to false. Will raise a
                SlicerNotConfigured error if true and the slicer has not been
                configured yet.
            require_exists (bool): Whether the profile is required to exist in
                order to be set as default. Defaults to true. Will raise a
                UnknownProfile error if true and the profile is unknown.

        Raises:
            ~octoprint.slicing.exceptions.UnknownSlicer: The slicer ``slicer``
                is unknown
            ~octoprint.slicing.exceptions.SlicerNotConfigured: The slicer ``slicer``
                has not yet been configured and ``require_configured`` was true.
            ~octoprint.slicing.exceptions.UnknownProfile: The profile ``name``
                was unknown for slicer ``slicer`` and ``require_exists`` was
                true.
        """
        if slicer not in self.registered_slicers:
            raise UnknownSlicer(slicer)
        if require_configured and slicer not in self.configured_slicers:
            raise SlicerNotConfigured(slicer)

        if not name:
            raise ValueError("name must be set")

        if require_exists and name not in self.all_profiles(
            slicer, require_configured=require_configured
        ):
            raise UnknownProfile(slicer, name)

        default_profiles = settings().get(["slicing", "defaultProfiles"])
        if not default_profiles:
            default_profiles = {}
        default_profiles[slicer] = name
        settings().set(["slicing", "defaultProfiles"], default_profiles)
        settings().save(force=True)

    def all_profiles(self, slicer, require_configured=False):
        """
        Retrieves all profiles for slicer ``slicer``.

        If ``require_configured`` is set to True (default is False), only will return the profiles if the ``slicer``
        is already configured, otherwise a :class:`SlicerNotConfigured` exception will be raised.

        Arguments:
            slicer (str): Identifier of the slicer for which to retrieve all slicer profiles
            require_configured (boolean): Whether to require the slicer ``slicer`` to be already configured (True)
                or not (False, default). If False and the slicer is not yet configured, a :class:`~octoprint.slicing.exceptions.SlicerNotConfigured`
                exception will be raised.

        Returns:
            dict of SlicingProfile: A dict of all :class:`SlicingProfile` instances available for the slicer ``slicer``, mapped by the identifier.

        Raises:
            ~octoprint.slicing.exceptions.UnknownSlicer: The slicer ``slicer`` is unknown.
            ~octoprint.slicing.exceptions.SlicerNotConfigured: The slicer ``slicer`` is not configured and ``require_configured`` was True.
        """

        if slicer not in self.registered_slicers:
            raise UnknownSlicer(slicer)
        if require_configured and slicer not in self.configured_slicers:
            raise SlicerNotConfigured(slicer)

        slicer_profile_path = self.get_slicer_profile_path(slicer)
        return self.get_slicer(slicer, require_configured=False).get_slicer_profiles(
            slicer_profile_path
        )

    def profiles_last_modified(self, slicer):
        """
        Retrieves the last modification date of ``slicer``'s profiles.

        Args:
            slicer (str): the slicer for which to retrieve the last modification date

        Returns:
            (float) the time stamp of the last modification of the slicer's profiles
        """

        if slicer not in self.registered_slicers:
            raise UnknownSlicer(slicer)

        slicer_profile_path = self.get_slicer_profile_path(slicer)
        return self.get_slicer(
            slicer, require_configured=False
        ).get_slicer_profiles_lastmodified(slicer_profile_path)

    def get_slicer_profile_path(self, slicer):
        """
        Retrieves the path where the profiles for slicer ``slicer`` are stored.

        Arguments:
            slicer (str): Identifier of the slicer for which to retrieve the path.

        Returns:
            str: The absolute path to the folder where the slicer's profiles are stored.

        Raises:
            ~octoprint.slicing.exceptions.UnknownSlicer: The slicer ``slicer`` is unknown.
        """

        if slicer not in self.registered_slicers:
            raise UnknownSlicer(slicer)

        path = os.path.join(self._profile_path, slicer)
        if not os.path.exists(path):
            os.makedirs(path)
        return path

    def get_profile_path(self, slicer, name, must_exist=False):
        """
        Retrieves the path to the profile named ``name`` for slicer ``slicer``.

        If ``must_exist`` is set to True (defaults to False) a :class:`UnknownProfile` exception will be raised if the
        profile doesn't exist yet.

        Arguments:
            slicer (str): Identifier of the slicer to which the profile belongs to.
            name (str): Identifier of the profile for which to retrieve the path.
            must_exist (boolean): Whether the path must exist (True) or not (False, default).

        Returns:
            str: The absolute path to the profile identified by ``name`` for slicer ``slicer``.

        Raises:
            ~octoprint.slicing.exceptions.UnknownSlicer: The slicer ``slicer`` is unknown.
            ~octoprint.slicing.exceptions.UnknownProfile: The profile named ``name`` doesn't exist and ``must_exist`` was True.
        """

        if slicer not in self.registered_slicers:
            raise UnknownSlicer(slicer)

        if not name:
            raise ValueError("name must be set")

        name = self._sanitize(name)

        path = os.path.join(
            self.get_slicer_profile_path(slicer), "{name}.profile".format(name=name)
        )
        if not os.path.realpath(path).startswith(os.path.realpath(self._profile_path)):
            raise IOError(
                "Path to profile {name} tried to break out of allows sub path".format(
                    **locals()
                )
            )
        if must_exist and not (os.path.exists(path) and os.path.isfile(path)):
            raise UnknownProfile(slicer, name)
        return path

    def _sanitize(self, name):
        if name is None:
            return None

        if "/" in name or "\\" in name:
            raise ValueError("name must not contain / or \\")

        import string

        valid_chars = "-_.() {ascii}{digits}".format(
            ascii=string.ascii_letters, digits=string.digits
        )
        sanitized_name = "".join(c for c in name if c in valid_chars)
        sanitized_name = sanitized_name.replace(" ", "_")
        return sanitized_name

    def _load_profile_from_path(self, slicer, path, require_configured=False):
        profile = self.get_slicer(
            slicer, require_configured=require_configured
        ).get_slicer_profile(path)
        default_profiles = settings().get(["slicing", "defaultProfiles"])
        if default_profiles and slicer in default_profiles:
            profile.default = default_profiles[slicer] == profile.name
        return profile

    def _save_profile_to_path(
        self,
        slicer,
        path,
        profile,
        allow_overwrite=True,
        overrides=None,
        require_configured=False,
    ):
        self.get_slicer(
            slicer, require_configured=require_configured
        ).save_slicer_profile(
            path, profile, allow_overwrite=allow_overwrite, overrides=overrides
        )

    def _get_default_profile(self, slicer):
        default_profiles = settings().get(["slicing", "defaultProfiles"])
        if default_profiles and slicer in default_profiles:
            try:
                return self.load_profile(slicer, default_profiles[slicer])
            except (UnknownProfile, IOError):
                # in that case we'll use the slicers predefined default profile
                pass

        return self.get_slicer(slicer).get_slicer_default_profile()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


class NoUpdateAvailable(Exception):
    pass


class UpdateAlreadyInProgress(Exception):
    pass


class UnknownUpdateType(Exception):
    pass


class UnknownCheckType(Exception):
    pass


class NetworkError(Exception):
    def __init__(self, message=None, cause=None):
        Exception.__init__(self)
        self.message = message
        self.cause = cause

    def __str__(self):
        if self.message is not None:
            return self.message
        elif self.cause is not None:
            return "NetworkError caused by {}".format(self.cause)
        else:
            return "NetworkError"


class CheckError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class RateLimitCheckError(CheckError):
    def __init__(self, message, remaining=None, limit=None, reset=None):
        super(RateLimitCheckError, self).__init__(message)
        self.remaining = remaining
        self.limit = limit
        self.reset = reset


class UpdateError(Exception):
    def __init__(self, message, data):
        self.message = message
        self.data = data

    def __str__(self):
        return self.message


class ScriptError(Exception):
    def __init__(self, returncode, stdout, stderr):
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr


class RestartFailed(Exception):
    pass


class ConfigurationInvalid(Exception):
    pass


class CannotCheckOffline(Exception):
    pass


class CannotUpdateOffline(Exception):
    pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2018 The OctoPrint Project - Released under terms of the AGPLv3 License"

import flask
from flask_babel import gettext

import octoprint.plugin
from octoprint.access import USER_GROUP
from octoprint.access.permissions import Permissions
from octoprint.events import Events


class Prompt(object):
    def __init__(self, text):
        self.text = text
        self.choices = []

        self._active = False

    @property
    def active(self):
        return self._active

    def add_choice(self, text):
        self.choices.append(text)

    def activate(self):
        self._active = True

    def validate_choice(self, choice):
        return 0 <= choice < len(self.choices)


class ActionCommandPromptPlugin(
    octoprint.plugin.AssetPlugin,
    octoprint.plugin.EventHandlerPlugin,
    octoprint.plugin.SettingsPlugin,
    octoprint.plugin.SimpleApiPlugin,
    octoprint.plugin.TemplatePlugin,
):

    COMMAND = "M876"
    CAP_PROMPT_SUPPORT = "PROMPT_SUPPORT"

    # noinspection PyMissingConstructor
    def __init__(self):
        self._prompt = None
        self._enable = "detected"
        self._command = None
        self._enable_emergency_sending = True
        self._enable_signal_support = True
        self._cap_prompt_support = False

    def initialize(self):
        self._enable = self._settings.get(["enable"])
        self._command = self._settings.get(["command"])
        self._enable_emergency_sending = self._settings.get_boolean(
            ["enable_emergency_sending"]
        )
        self._enable_signal_support = self._settings.get_boolean(
            ["enable_signal_support"]
        )

    # Additional permissions hook

    def get_additional_permissions(self):
        return [
            {
                "key": "INTERACT",
                "name": "Interact with printer prompts",
                "description": gettext("Allows to see and interact with printer prompts"),
                "default_groups": [USER_GROUP],
                "roles": ["interact"],
            }
        ]

    # ~ AssetPlugin

    def get_assets(self):
        return {
            "js": ["js/action_command_prompt.js"],
            "clientjs": ["clientjs/action_command_prompt.js"],
        }

    # ~ EventHandlerPlugin

    def on_event(self, event, payload):
        if (
            event == Events.CONNECTED
            and self._enable == "always"
            and self._enable_signal_support
        ):
            self._printer.commands(["{command} P1".format(command=self._command)])
        elif event == Events.DISCONNECTED:
            self._close_prompt()

    # ~ SettingsPlugin

    def get_settings_defaults(self):
        return {
            "enable": "detected",
            "command": self.COMMAND,
            "enable_emergency_sending": True,
            "enable_signal_support": True,
        }

    def on_settings_save(self, data):
        octoprint.plugin.SettingsPlugin.on_settings_save(self, data)
        self._enable = self._settings.get(["enable"])
        self._command = self._settings.get(["command"])
        self._enable_emergency_sending = self._settings.get_boolean(
            ["enable_emergency_sending"]
        )
        self._enable_signal_support = self._settings.get_boolean(
            ["enable_signal_support"]
        )

    # ~ SimpleApiPlugin

    def get_api_commands(self):
        return {"select": ["choice"]}

    def on_api_command(self, command, data):
        if command == "select":
            if not Permissions.PLUGIN_ACTION_COMMAND_PROMPT_INTERACT.can():
                return flask.abort(403, "Insufficient permissions")

            if self._prompt is None:
                return flask.abort(409, "No active prompt")

            choice = data["choice"]
            if not isinstance(choice, int) or not self._prompt.validate_choice(choice):
                return flask.abort(
                    400, "{!r} is not a valid value for choice".format(choice)
                )

            self._answer_prompt(choice)

    def on_api_get(self, request):
        if not Permissions.PLUGIN_ACTION_COMMAND_PROMPT_INTERACT.can():
            return flask.abort(403, "Insufficient permissions")
        if self._prompt is None:
            return flask.jsonify()
        else:
            return flask.jsonify(text=self._prompt.text, choices=self._prompt.choices)

    # ~ TemplatePlugin

    def get_template_configs(self):
        return [
            {
                "type": "settings",
                "name": gettext("Printer Dialogs"),
                "custom_bindings": False,
            }
        ]

    # ~ action command handler

    def action_command_handler(self, comm, line, action, *args, **kwargs):
        if not action.startswith("prompt_"):
            return

        parts = action.split(None, 1)
        if len(parts) == 1:
            action = parts[0]
            parameter = ""
        else:
            action, parameter = parts

        if action == "prompt_begin":
            if self._prompt is not None and self._prompt.active:
                self._logger.warning("Prompt is already defined")
                return
            self._prompt = Prompt(parameter.strip())

        elif action == "prompt_choice" or action == "prompt_button":
            if self._prompt is None:
                return
            if self._prompt.active:
                self._logger.warning("Prompt is already active")
                return
            self._prompt.add_choice(parameter.strip())

        elif action == "prompt_show":
            if self._prompt is None:
                return
            if self._prompt.active:
                self._logger.warning("Prompt is already active")
                return
            self._show_prompt()

        elif action == "prompt_end":
            if self._prompt is None:
                return
            self._close_prompt()
            self._prompt = None

    # ~ queuing handling

    def gcode_queuing_handler(
        self,
        comm_instance,
        phase,
        cmd,
        cmd_type,
        gcode,
        subcode=None,
        tags=None,
        *args,
        **kwargs
    ):
        if gcode != self._command:
            return

        if (
            self._enable == "never"
            or (self._enable == "detected" and not self._cap_prompt_support)
            or not self._enable_emergency_sending
        ):
            return

        if "S" not in cmd:
            # we only force-send M876 Sx
            return

        # noinspection PyProtectedMember
        return comm_instance._emergency_force_send(
            cmd, "Force-sending {} to the printer".format(self._command), gcode=gcode
        )

    # ~ capability reporting

    def firmware_capability_handler(
        self, comm_instance, capability, enabled, already_defined, *args, **kwargs
    ):
        if capability == self.CAP_PROMPT_SUPPORT and enabled:
            self._cap_prompt_support = True
            if self._enable == "detected" and self._enable_signal_support:
                self._printer.commands(["{command} P1".format(command=self._command)])

    # ~ prompt handling

    def _show_prompt(self):
        if self._enable == "never" or (
            self._enable == "detected" and not self._cap_prompt_support
        ):
            return

        self._prompt.activate()
        self._plugin_manager.send_plugin_message(
            self._identifier,
            {
                "action": "show",
                "text": self._prompt.text,
                "choices": self._prompt.choices,
            },
        )

    def _close_prompt(self):
        if self._enable == "never" or (
            self._enable == "detected" and not self._cap_prompt_support
        ):
            return

        self._prompt = None
        self._plugin_manager.send_plugin_message(self._identifier, {"action": "close"})

    def _answer_prompt(self, choice):
        if self._enable == "never" or (
            self._enable == "detected" and not self._cap_prompt_support
        ):
            return

        self._close_prompt()
        if "{choice}" in self._command:
            self._printer.commands([self._command.format(choice=choice)], force=True)
        else:
            self._printer.commands(
                ["{command} S{choice}".format(command=self._command, choice=choice)],
                force=True,
            )


__plugin_name__ = "Action Command Prompt Support"
__plugin_description__ = (
    "Allows your printer to trigger prompts via action commands on the connection"
)
__plugin_author__ = "Gina Häußge"
__plugin_disabling_discouraged__ = gettext(
    "Without this plugin your printer will no longer be able to trigger"
    " confirmation or selection prompts in OctoPrint"
)
__plugin_license__ = "AGPLv3"
__plugin_pythoncompat__ = ">=2.7,<4"
__plugin_implementation__ = ActionCommandPromptPlugin()
__plugin_hooks__ = {
    "octoprint.comm.protocol.action": __plugin_implementation__.action_command_handler,
    "octoprint.comm.protocol.gcode.queuing": __plugin_implementation__.gcode_queuing_handler,
    "octoprint.comm.protocol.firmware.capabilities": __plugin_implementation__.firmware_capability_handler,
    "octoprint.access.permissions": __plugin_implementation__.get_additional_permissions,
}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"

import datetime
import fnmatch
import io
import logging
import os
import shutil
import sys
import threading
import time

try:
    import queue
except ImportError:
    import Queue as queue

import collections
import re

import requests
import sarge

import octoprint.plugin
import octoprint.util as util
from octoprint.events import Events, eventManager
from octoprint.plugin import plugin_manager
from octoprint.settings import settings
from octoprint.util import get_fully_qualified_classname as fqcn
from octoprint.util import monotonic_time, sv
from octoprint.util.commandline import CommandlineCaller

try:
    from os import scandir
except ImportError:
    from scandir import scandir


# currently configured timelapse
current = None

# currently active render job, if any
current_render_job = None

# filename formats
_capture_format = "{prefix}-%d.jpg"
_capture_glob = "{prefix}-*.jpg"
_output_format = "{prefix}{postfix}.{extension}"

# ffmpeg progress regexes
_ffmpeg_duration_regex = re.compile(r"Duration: (\d{2}):(\d{2}):(\d{2})\.\d{2}")
_ffmpeg_current_regex = re.compile(r"time=(\d{2}):(\d{2}):(\d{2})\.\d{2}")

# old capture format, needed to delete old left-overs from
# versions <1.2.9
_old_capture_format_re = re.compile(r"^tmp_\d{5}.jpg$")

# valid timelapses
_valid_timelapse_types = ["off", "timed", "zchange"]

# callbacks for timelapse config updates
_update_callbacks = []

# lock for timelapse cleanup, must be re-entrant
_cleanup_lock = threading.RLock()

# lock for timelapse job
_job_lock = threading.RLock()

# cached valid timelapse extensions
_extensions = None


def valid_timelapse(path):
    global _extensions

    if _extensions is None:
        # create list of extensions
        extensions = ["mpg", "mpeg", "mp4", "m4v", "mkv"]

        hooks = plugin_manager().get_hooks("octoprint.timelapse.extensions")
        for name, hook in hooks.items():
            try:
                result = hook()
                if result is None or not isinstance(result, list):
                    continue
                extensions += result
            except Exception:
                logging.getLogger(__name__).exception(
                    "Exception while retrieving additional timelapse "
                    "extensions from hook {name}".format(name=name),
                    extra={"plugin": name},
                )

        _extensions = list(set(extensions))

    return util.is_allowed_file(path, _extensions)


def _extract_prefix(filename):
    """
    >>> _extract_prefix("some_long_filename_without_hyphen.jpg")
    >>> _extract_prefix("-first_char_is_hyphen.jpg")
    >>> _extract_prefix("some_long_filename_with-stuff.jpg") # doctest: +ALLOW_UNICODE
    'some_long_filename_with'
    """
    pos = filename.rfind("-")
    if not pos or pos < 0:
        return None
    return filename[:pos]


def last_modified_finished():
    return os.stat(settings().getBaseFolder("timelapse", check_writable=False)).st_mtime


def last_modified_unrendered():
    return os.stat(
        settings().getBaseFolder("timelapse_tmp", check_writable=False)
    ).st_mtime


def get_finished_timelapses():
    files = []
    basedir = settings().getBaseFolder("timelapse", check_writable=False)
    for entry in scandir(basedir):
        if util.is_hidden_path(entry.path) or not valid_timelapse(entry.path):
            continue
        files.append(
            {
                "name": entry.name,
                "size": util.get_formatted_size(entry.stat().st_size),
                "bytes": entry.stat().st_size,
                "date": util.get_formatted_datetime(
                    datetime.datetime.fromtimestamp(entry.stat().st_mtime)
                ),
            }
        )
    return files


def get_unrendered_timelapses():
    global _job_lock
    global current

    delete_old_unrendered_timelapses()

    basedir = settings().getBaseFolder("timelapse_tmp", check_writable=False)
    jobs = collections.defaultdict(
        lambda: {"count": 0, "size": None, "bytes": 0, "date": None, "timestamp": None}
    )

    for entry in scandir(basedir):
        if not fnmatch.fnmatch(entry.name, "*.jpg"):
            continue

        prefix = _extract_prefix(entry.name)
        if prefix is None:
            continue

        jobs[prefix]["count"] += 1
        jobs[prefix]["bytes"] += entry.stat().st_size
        if (
            jobs[prefix]["timestamp"] is None
            or entry.stat().st_mtime < jobs[prefix]["timestamp"]
        ):
            jobs[prefix]["timestamp"] = entry.stat().st_mtime

    with _job_lock:
        global current_render_job

        def finalize_fields(prefix, job):
            currently_recording = current is not None and current.prefix == prefix
            currently_rendering = (
                current_render_job is not None and current_render_job["prefix"] == prefix
            )

            job["size"] = util.get_formatted_size(job["bytes"])
            job["date"] = util.get_formatted_datetime(
                datetime.datetime.fromtimestamp(job["timestamp"])
            )
            job["recording"] = currently_recording
            job["rendering"] = currently_rendering
            job["processing"] = currently_recording or currently_rendering
            del job["timestamp"]

            return job

        return sorted(
            [
                util.dict_merge({"name": key}, finalize_fields(key, value))
                for key, value in jobs.items()
            ],
            key=lambda x: sv(x["name"]),
        )


def delete_unrendered_timelapse(name):
    global _cleanup_lock

    pattern = "{}*.jpg".format(util.glob_escape(name))

    basedir = settings().getBaseFolder("timelapse_tmp")
    with _cleanup_lock:
        for entry in scandir(basedir):
            try:
                if fnmatch.fnmatch(entry.name, pattern):
                    os.remove(entry.path)
            except Exception:
                if logging.getLogger(__name__).isEnabledFor(logging.DEBUG):
                    logging.getLogger(__name__).exception(
                        "Error while processing file {} during cleanup".format(entry.name)
                    )


def render_unrendered_timelapse(name, gcode=None, postfix=None, fps=None):
    capture_dir = settings().getBaseFolder("timelapse_tmp")
    output_dir = settings().getBaseFolder("timelapse")

    if fps is None:
        fps = settings().getInt(["webcam", "timelapse", "fps"])
    threads = settings().get(["webcam", "ffmpegThreads"])
    videocodec = settings().get(["webcam", "ffmpegVideoCodec"])

    job = TimelapseRenderJob(
        capture_dir,
        output_dir,
        name,
        postfix=postfix,
        fps=fps,
        threads=threads,
        videocodec=videocodec,
        on_start=_create_render_start_handler(name, gcode=gcode),
        on_success=_create_render_success_handler(name, gcode=gcode),
        on_fail=_create_render_fail_handler(name, gcode=gcode),
        on_always=_create_render_always_handler(name, gcode=gcode),
    )
    job.process()


def delete_old_unrendered_timelapses():
    global _cleanup_lock

    basedir = settings().getBaseFolder("timelapse_tmp")
    clean_after_days = settings().getInt(["webcam", "cleanTmpAfterDays"])
    cutoff = time.time() - clean_after_days * 24 * 60 * 60

    prefixes_to_clean = []

    with _cleanup_lock:
        for entry in scandir(basedir):
            try:
                prefix = _extract_prefix(entry.name)
                if prefix is None:
                    # might be an old tmp_00000.jpg kinda frame. we can't
                    # render those easily anymore, so delete that stuff
                    if _old_capture_format_re.match(entry.name):
                        os.remove(entry.path)
                    continue

                if prefix in prefixes_to_clean:
                    continue

                # delete if both creation and modification time are older than the cutoff
                if max(entry.stat().st_ctime, entry.stat().st_mtime) < cutoff:
                    prefixes_to_clean.append(prefix)
            except Exception:
                if logging.getLogger(__name__).isEnabledFor(logging.DEBUG):
                    logging.getLogger(__name__).exception(
                        "Error while processing file {} during cleanup".format(entry.name)
                    )

        for prefix in prefixes_to_clean:
            delete_unrendered_timelapse(prefix)
            logging.getLogger(__name__).info(
                "Deleted old unrendered timelapse {}".format(prefix)
            )


def _create_render_start_handler(name, gcode=None):
    def f(movie):
        global _job_lock

        with _job_lock:
            global current_render_job
            payload = {
                "gcode": gcode if gcode is not None else "unknown",
                "movie": movie,
                "movie_basename": os.path.basename(movie),
                "movie_prefix": name,
            }
            current_render_job = {"prefix": name}
            current_render_job.update(payload)
        eventManager().fire(Events.MOVIE_RENDERING, payload)

    return f


def _create_render_success_handler(name, gcode=None):
    def f(movie):
        delete_unrendered_timelapse(name)
        payload = {
            "gcode": gcode if gcode is not None else "unknown",
            "movie": movie,
            "movie_basename": os.path.basename(movie),
            "movie_prefix": name,
        }
        eventManager().fire(Events.MOVIE_DONE, payload)

    return f


def _create_render_fail_handler(name, gcode=None):
    def f(
        movie,
        returncode=255,
        stdout="Unknown error",
        stderr="Unknown error",
        reason="unknown",
    ):
        payload = {
            "gcode": gcode if gcode is not None else "unknown",
            "movie": movie,
            "movie_basename": os.path.basename(movie),
            "movie_prefix": name,
            "returncode": returncode,
            "out": stdout,
            "error": stderr,
            "reason": reason,
        }
        eventManager().fire(Events.MOVIE_FAILED, payload)

    return f


def _create_render_always_handler(name, gcode=None):
    def f(movie):
        global current_render_job
        global _job_lock
        with _job_lock:
            current_render_job = None

    return f


def register_callback(callback):
    if callback not in _update_callbacks:
        _update_callbacks.append(callback)


def unregister_callback(callback):
    try:
        _update_callbacks.remove(callback)
    except ValueError:
        # not registered
        pass


def notify_callbacks(timelapse):
    if timelapse is None:
        config = None
    else:
        config = timelapse.config_data()

    for callback in _update_callbacks:
        notify_callback(callback, config)


def notify_callback(callback, config=None, timelapse=None):
    if config is None and timelapse is not None:
        config = timelapse.config_data()

    try:
        callback.sendTimelapseConfig(config)
    except Exception:
        logging.getLogger(__name__).exception(
            "Exception while pushing timelapse configuration",
            extra={"callback": fqcn(callback)},
        )


def configure_timelapse(config=None, persist=False):
    global current

    if config is None:
        config = settings().get(["webcam", "timelapse"], merged=True)

    if current is not None:
        current.unload()

    snapshot_url = settings().get(["webcam", "snapshot"])
    ffmpeg_path = settings().get(["webcam", "ffmpeg"])
    timelapse_enabled = settings().getBoolean(["webcam", "timelapseEnabled"])
    timelapse_precondition = (
        snapshot_url is not None
        and snapshot_url.strip() != ""
        and ffmpeg_path is not None
        and ffmpeg_path.strip() != ""
    )

    type = config["type"]
    if not timelapse_precondition and timelapse_precondition:
        logging.getLogger(__name__).warning(
            "Essential timelapse settings unconfigured (snapshot URL or FFMPEG path) "
            "but timelapse enabled."
        )

    if (
        not timelapse_enabled
        or not timelapse_precondition
        or type is None
        or "off" == type
    ):
        current = None

    else:
        postRoll = 0
        if "postRoll" in config and config["postRoll"] >= 0:
            postRoll = config["postRoll"]

        fps = 25
        if "fps" in config and config["fps"] > 0:
            fps = config["fps"]

        if "zchange" == type:
            retractionZHop = 0
            if (
                "options" in config
                and "retractionZHop" in config["options"]
                and config["options"]["retractionZHop"] >= 0
            ):
                retractionZHop = config["options"]["retractionZHop"]

            minDelay = 5
            if (
                "options" in config
                and "minDelay" in config["options"]
                and config["options"]["minDelay"] > 0
            ):
                minDelay = config["options"]["minDelay"]

            current = ZTimelapse(
                post_roll=postRoll,
                retraction_zhop=retractionZHop,
                min_delay=minDelay,
                fps=fps,
            )

        elif "timed" == type:
            interval = 10
            if (
                "options" in config
                and "interval" in config["options"]
                and config["options"]["interval"] > 0
            ):
                interval = config["options"]["interval"]

            current = TimedTimelapse(post_roll=postRoll, interval=interval, fps=fps)

    notify_callbacks(current)

    if persist:
        settings().set(["webcam", "timelapse"], config)
        settings().save()


class Timelapse(object):
    QUEUE_ENTRY_TYPE_CAPTURE = "capture"
    QUEUE_ENTRY_TYPE_CALLBACK = "callback"

    def __init__(self, post_roll=0, fps=25):
        self._logger = logging.getLogger(__name__)
        self._image_number = None
        self._in_timelapse = False
        self._gcode_file = None
        self._file_prefix = None

        self._capture_errors = 0
        self._capture_success = 0

        self._post_roll = post_roll
        self._on_post_roll_done = None

        self._capture_dir = settings().getBaseFolder("timelapse_tmp")
        self._movie_dir = settings().getBaseFolder("timelapse")
        self._snapshot_url = settings().get(["webcam", "snapshot"])
        self._snapshot_timeout = settings().getInt(["webcam", "snapshotTimeout"])
        self._snapshot_validate_ssl = settings().getBoolean(
            ["webcam", "snapshotSslValidation"]
        )

        self._fps = fps

        self._pluginManager = octoprint.plugin.plugin_manager()
        self._pre_capture_hooks = self._pluginManager.get_hooks(
            "octoprint.timelapse.capture.pre"
        )
        self._post_capture_hooks = self._pluginManager.get_hooks(
            "octoprint.timelapse.capture.post"
        )

        self._capture_mutex = threading.Lock()
        self._capture_queue = queue.Queue()
        self._capture_queue_active = True

        self._capture_queue_thread = threading.Thread(target=self._capture_queue_worker)
        self._capture_queue_thread.daemon = True
        self._capture_queue_thread.start()

        # subscribe events
        eventManager().subscribe(Events.PRINT_STARTED, self.on_print_started)
        eventManager().subscribe(Events.PRINT_FAILED, self.on_print_done)
        eventManager().subscribe(Events.PRINT_DONE, self.on_print_done)
        eventManager().subscribe(Events.PRINT_RESUMED, self.on_print_resumed)
        for (event, callback) in self.event_subscriptions():
            eventManager().subscribe(event, callback)

    @property
    def prefix(self):
        return self._file_prefix

    @property
    def post_roll(self):
        return self._post_roll

    @property
    def fps(self):
        return self._fps

    def unload(self):
        if self._in_timelapse:
            self.stop_timelapse(do_create_movie=False)

        # unsubscribe events
        eventManager().unsubscribe(Events.PRINT_STARTED, self.on_print_started)
        eventManager().unsubscribe(Events.PRINT_FAILED, self.on_print_done)
        eventManager().unsubscribe(Events.PRINT_DONE, self.on_print_done)
        eventManager().unsubscribe(Events.PRINT_RESUMED, self.on_print_resumed)
        for (event, callback) in self.event_subscriptions():
            eventManager().unsubscribe(event, callback)

    def on_print_started(self, event, payload):
        """
        Override this to perform additional actions upon start of a print job.
        """
        self.start_timelapse(payload["name"])

    def on_print_done(self, event, payload):
        """
        Override this to perform additional actions upon the stop of a print job.
        """
        self.stop_timelapse(success=(event == Events.PRINT_DONE))

    def on_print_resumed(self, event, payload):
        """
        Override this to perform additional actions upon the pausing of a print job.
        """
        if not self._in_timelapse:
            self.start_timelapse(payload["name"])

    def event_subscriptions(self):
        """
        Override this method to subscribe to additional events by returning an array of (event, callback) tuples.

        Events that are already subscribed:
          * PrintStarted - self.onPrintStarted
          * PrintResumed - self.onPrintResumed
          * PrintFailed - self.onPrintDone
          * PrintDone - self.onPrintDone
        """
        return []

    def config_data(self):
        """
        Override this method to return the current timelapse configuration data. The data should have the following
        form:

            type: "<type of timelapse>",
            options: { <additional options> }
        """
        return None

    def start_timelapse(self, gcode_file):
        self._logger.debug("Starting timelapse for %s" % gcode_file)

        self._image_number = 0
        self._capture_errors = 0
        self._capture_success = 0
        self._in_timelapse = True
        self._gcode_file = os.path.basename(gcode_file)
        self._file_prefix = "{}_{}".format(
            os.path.splitext(self._gcode_file)[0], time.strftime("%Y%m%d%H%M%S")
        )

    def stop_timelapse(self, do_create_movie=True, success=True):
        self._logger.debug("Stopping timelapse")

        self._in_timelapse = False

        def reset_image_number():
            self._image_number = None

        def create_movie():
            render_unrendered_timelapse(
                self._file_prefix,
                gcode=self._gcode_file,
                postfix=None if success else "-fail",
                fps=self._fps,
            )

        def reset_and_create():
            reset_image_number()
            create_movie()

        def wait_for_captures(callback):
            self._capture_queue.put(
                {"type": self.__class__.QUEUE_ENTRY_TYPE_CALLBACK, "callback": callback}
            )

        def create_wait_for_captures(callback):
            def f():
                wait_for_captures(callback)

            return f

        # wait for everything so far in the queue to be processed, then see if we should process from there
        def continue_rendering():
            if self._capture_success == 0:
                # no images - either nothing was attempted to be captured or all attempts ran into an error
                if self._capture_errors > 0:
                    # this is the latter case
                    _create_render_fail_handler(
                        self._file_prefix, gcode=self._gcode_file
                    )("n/a", returncode=0, stdout="", stderr="", reason="no_frames")

                # in any case, don't continue
                return

            # check if we have post roll configured
            if self._post_roll > 0:
                # capture post roll, wait for THAT to finish, THEN render
                eventManager().fire(
                    Events.POSTROLL_START,
                    {
                        "postroll_duration": self.calculate_post_roll(),
                        "postroll_length": self.post_roll,
                        "postroll_fps": self.fps,
                    },
                )
                if do_create_movie:
                    self._on_post_roll_done = create_wait_for_captures(reset_and_create)
                else:
                    self._on_post_roll_done = reset_image_number
                self.process_post_roll()
            else:
                # no post roll? perfect, render
                if do_create_movie:
                    wait_for_captures(reset_and_create)
                else:
                    reset_image_number()

        self._logger.debug("Waiting to process capture queue")
        wait_for_captures(continue_rendering)

    def calculate_post_roll(self):
        return None

    def process_post_roll(self):
        self.post_roll_finished()

    def post_roll_finished(self):
        if self.post_roll:
            eventManager().fire(Events.POSTROLL_END)
            if self._on_post_roll_done is not None:
                self._on_post_roll_done()

    def capture_image(self):
        if self._capture_dir is None:
            self._logger.warning("Cannot capture image, capture directory is unset")
            return

        with self._capture_mutex:
            if self._image_number is None:
                self._logger.warning("Cannot capture image, image number is unset")
                return

            filename = os.path.join(
                self._capture_dir,
                _capture_format.format(prefix=self._file_prefix) % self._image_number,
            )
            self._image_number += 1

        self._logger.debug("Capturing image to {}".format(filename))
        entry = {
            "type": self.__class__.QUEUE_ENTRY_TYPE_CAPTURE,
            "filename": filename,
            "onerror": self._on_capture_error,
        }
        self._capture_queue.put(entry)
        return filename

    def _on_capture_error(self):
        with self._capture_mutex:
            if self._image_number is not None and self._image_number > 0:
                self._image_number -= 1

    def _capture_queue_worker(self):
        while self._capture_queue_active:
            entry = self._capture_queue.get(block=True)

            if (
                entry["type"] == self.__class__.QUEUE_ENTRY_TYPE_CAPTURE
                and "filename" in entry
            ):
                filename = entry["filename"]
                onerror = entry.pop("onerror", None)
                self._perform_capture(filename, onerror=onerror)

            elif (
                entry["type"] == self.__class__.QUEUE_ENTRY_TYPE_CALLBACK
                and "callback" in entry
            ):
                args = entry.pop("args", [])
                kwargs = entry.pop("kwargs", {})
                entry["callback"](*args, **kwargs)

    def _perform_capture(self, filename, onerror=None):
        # pre-capture hook
        for hook in self._pre_capture_hooks.values():
            try:
                hook(filename)
            except Exception:
                self._logger.exception(
                    "Error while processing hook {name}.".format(**locals())
                )

        eventManager().fire(Events.CAPTURE_START, {"file": filename})
        try:
            self._logger.debug(
                "Going to capture {} from {}".format(filename, self._snapshot_url)
            )
            r = requests.get(
                self._snapshot_url,
                stream=True,
                timeout=self._snapshot_timeout,
                verify=self._snapshot_validate_ssl,
            )
            r.raise_for_status()

            with io.open(filename, "wb") as f:
                for chunk in r.iter_content(chunk_size=1024):
                    if chunk:
                        f.write(chunk)
                        f.flush()

            self._logger.debug(
                "Image {} captured from {}".format(filename, self._snapshot_url)
            )
        except Exception as e:
            self._logger.exception(
                "Could not capture image {} from {}".format(filename, self._snapshot_url)
            )
            self._capture_errors += 1
            err = e
        else:
            self._capture_success += 1
            err = None

        # post-capture hook
        for hook in self._post_capture_hooks.values():
            try:
                hook(filename, err is None)
            except Exception:
                self._logger.exception(
                    "Error while processing hook {name}.".format(**locals())
                )

        # handle events and onerror call
        if err is None:
            eventManager().fire(Events.CAPTURE_DONE, {"file": filename})
            return True
        else:
            if callable(onerror):
                onerror()
            eventManager().fire(
                Events.CAPTURE_FAILED,
                {"file": filename, "error": str(err), "url": self._snapshot_url},
            )
            return False

    def _copying_postroll(self):
        with self._capture_mutex:
            filename = os.path.join(
                self._capture_dir,
                _capture_format.format(prefix=self._file_prefix) % self._image_number,
            )
            self._image_number += 1

        if self._perform_capture(filename):
            for _ in range(self._post_roll * self._fps):
                newFile = os.path.join(
                    self._capture_dir,
                    _capture_format.format(prefix=self._file_prefix) % self._image_number,
                )
                self._image_number += 1
                shutil.copyfile(filename, newFile)

    def clean_capture_dir(self):
        if not os.path.isdir(self._capture_dir):
            self._logger.warning("Cannot clean capture directory, it is unset")
            return
        delete_unrendered_timelapse(self._file_prefix)


class ZTimelapse(Timelapse):
    def __init__(self, retraction_zhop=0, min_delay=5.0, post_roll=0, fps=25):
        Timelapse.__init__(self, post_roll=post_roll, fps=fps)

        if min_delay < 0:
            min_delay = 0

        self._retraction_zhop = retraction_zhop
        self._min_delay = min_delay
        self._last_snapshot = None
        self._logger.debug("ZTimelapse initialized")

    @property
    def retraction_zhop(self):
        return self._retraction_zhop

    @property
    def min_delay(self):
        return self._min_delay

    def event_subscriptions(self):
        return [(Events.Z_CHANGE, self._on_z_change)]

    def config_data(self):
        return {"type": "zchange", "options": {"retractionZHop": self._retraction_zhop}}

    def process_post_roll(self):
        # we always copy the final image for the whole post roll
        # for z based timelapses
        self._copying_postroll()
        Timelapse.process_post_roll(self)

    def _on_z_change(self, event, payload):
        # check if height difference equals z-hop, if so don't take a picture
        if (
            self._retraction_zhop != 0
            and payload["old"] is not None
            and payload["new"] is not None
        ):
            diff = round(abs(payload["new"] - payload["old"]), 3)
            zhop = round(self._retraction_zhop, 3)
            if diff == zhop:
                return

        # check if last picture has been less than min_delay ago, if so don't take a picture (anti vase mode...)
        now = monotonic_time()
        if (
            self._min_delay
            and self._last_snapshot
            and self._last_snapshot + self._min_delay > now
        ):
            self._logger.debug("Rate limited z-change, not taking a snapshot")
            return

        self.capture_image()
        self._last_snapshot = now


class TimedTimelapse(Timelapse):
    def __init__(self, interval=1, post_roll=0, fps=25):
        Timelapse.__init__(self, post_roll=post_roll, fps=fps)
        self._interval = interval
        if self._interval < 1:
            self._interval = 1  # force minimum interval of 1s
        self._timer = None
        self._logger.debug("TimedTimelapse initialized")

    @property
    def interval(self):
        return self._interval

    def config_data(self):
        return {"type": "timed", "options": {"interval": self._interval}}

    def on_print_started(self, event, payload):
        Timelapse.on_print_started(self, event, payload)
        if self._timer is not None:
            return

        self._logger.debug("Starting timer for interval based timelapse")
        from octoprint.util import RepeatedTimer

        self._timer = RepeatedTimer(
            self._interval,
            self._timer_task,
            run_first=True,
            condition=self._timer_active,
            on_finish=self._on_timer_finished,
        )
        self._timer.start()

    def process_post_roll(self):
        # we only use the final image as post roll
        self._copying_postroll()
        self.post_roll_finished()

    def _timer_active(self):
        return self._in_timelapse

    def _timer_task(self):
        self.capture_image()

    def _on_timer_finished(self):
        # timer is done, delete it
        self._timer = None


class TimelapseRenderJob(object):

    render_job_lock = threading.RLock()

    def __init__(
        self,
        capture_dir,
        output_dir,
        prefix,
        postfix=None,
        capture_glob=_capture_glob,
        capture_format=_capture_format,
        output_format=_output_format,
        fps=25,
        threads=1,
        videocodec="mpeg2video",
        on_start=None,
        on_success=None,
        on_fail=None,
        on_always=None,
    ):
        self._capture_dir = capture_dir
        self._output_dir = output_dir
        self._prefix = prefix
        self._postfix = postfix
        self._capture_glob = capture_glob
        self._capture_format = capture_format
        self._output_format = output_format
        self._fps = fps
        self._threads = threads
        self._videocodec = videocodec
        self._on_start = on_start
        self._on_success = on_success
        self._on_fail = on_fail
        self._on_always = on_always

        self._thread = None
        self._logger = logging.getLogger(__name__)

        self._parsed_duration = 0

    def process(self):
        """Processes the job."""

        self._thread = threading.Thread(
            target=self._render,
            name="TimelapseRenderJob_{prefix}_{postfix}".format(
                prefix=self._prefix, postfix=self._postfix
            ),
        )
        self._thread.daemon = True
        self._thread.start()

    def _render(self):
        """Rendering runnable."""

        ffmpeg = settings().get(["webcam", "ffmpeg"])
        bitrate = settings().get(["webcam", "bitrate"])
        if ffmpeg is None or bitrate is None:
            self._logger.warning(
                "Cannot create movie, path to ffmpeg or desired bitrate is unset"
            )
            return

        if self._videocodec == "mpeg2video":
            extension = "mpg"
        else:
            extension = "mp4"

        input = os.path.join(
            self._capture_dir,
            self._capture_format.format(
                prefix=self._prefix,
                postfix=self._postfix if self._postfix is not None else "",
            ),
        )

        output_name = self._output_format.format(
            prefix=self._prefix,
            postfix=self._postfix if self._postfix is not None else "",
            extension=extension,
        )
        temporary = os.path.join(self._output_dir, ".{}".format(output_name))
        output = os.path.join(self._output_dir, output_name)

        for i in range(4):
            if os.path.exists(input % i):
                break
        else:
            self._logger.warning("Cannot create a movie, no frames captured")
            self._notify_callback(
                "fail", output, returncode=0, stdout="", stderr="", reason="no_frames"
            )
            return

        hflip = settings().getBoolean(["webcam", "flipH"])
        vflip = settings().getBoolean(["webcam", "flipV"])
        rotate = settings().getBoolean(["webcam", "rotate90"])

        watermark = None
        if settings().getBoolean(["webcam", "watermark"]):
            watermark = os.path.join(
                os.path.dirname(__file__), "static", "img", "watermark.png"
            )
            if sys.platform == "win32":
                # Because ffmpeg hiccups on windows' drive letters and backslashes we have to give the watermark
                # path a special treatment. Yeah, I couldn't believe it either...
                watermark = watermark.replace("\\", "/").replace(":", "\\\\:")

        # prepare ffmpeg command
        command_str = self._create_ffmpeg_command_string(
            ffmpeg,
            self._fps,
            bitrate,
            self._threads,
            input,
            temporary,
            self._videocodec,
            hflip=hflip,
            vflip=vflip,
            rotate=rotate,
            watermark=watermark,
        )
        self._logger.debug("Executing command: {}".format(command_str))

        with self.render_job_lock:
            try:
                self._notify_callback("start", output)

                self._logger.debug("Parsing ffmpeg output")

                c = CommandlineCaller()
                c.on_log_stderr = self._process_ffmpeg_output
                returncode, stdout_text, stderr_text = c.call(
                    command_str, delimiter=b"\r", buffer_size=512
                )

                self._logger.debug("Done with parsing")

                if returncode == 0:
                    shutil.move(temporary, output)
                    self._notify_callback("success", output)
                else:
                    self._logger.warning(
                        "Could not render movie, got return code %r: %s"
                        % (returncode, stderr_text)
                    )
                    self._notify_callback(
                        "fail",
                        output,
                        returncode=returncode,
                        stdout=stdout_text,
                        stderr=stderr_text,
                        reason="returncode",
                    )
            except Exception:
                self._logger.exception("Could not render movie due to unknown error")
                self._notify_callback("fail", output, reason="unknown")
            finally:
                try:
                    if os.path.exists(temporary):
                        os.remove(temporary)
                except Exception:
                    self._logger.warning(
                        "Could not delete temporary timelapse {}".format(temporary)
                    )
                self._notify_callback("always", output)

    def _process_ffmpeg_output(self, *lines):
        for line in lines:
            # We should be getting the time more often, so try it first
            current_time = _ffmpeg_current_regex.search(line)
            if current_time is not None and self._parsed_duration != 0:
                current_s = self._convert_time(*current_time.groups())
                progress = current_s / float(self._parsed_duration) * 100

                # Update progress bar
                for callback in _update_callbacks:
                    try:
                        callback.sendRenderProgress(progress)
                    except Exception:
                        self._logger.exception("Exception while pushing render progress")

            else:
                duration = _ffmpeg_duration_regex.search(line)
                if duration is not None:
                    self._parsed_duration = self._convert_time(*duration.groups())

    @staticmethod
    def _convert_time(hours, minutes, seconds):
        return (int(hours) * 60 + int(minutes)) * 60 + int(seconds)

    @classmethod
    def _create_ffmpeg_command_string(
        cls,
        ffmpeg,
        fps,
        bitrate,
        threads,
        input,
        output,
        videocodec,
        hflip=False,
        vflip=False,
        rotate=False,
        watermark=None,
        pixfmt="yuv420p",
    ):
        """
        Create ffmpeg command string based on input parameters.

        Arguments:
            ffmpeg (str): Path to ffmpeg
            fps (int): Frames per second for output
            bitrate (str): Bitrate of output
            threads (int): Number of threads to use for rendering
            videocodec (str): Videocodec to be used for encoding
            input (str): Absolute path to input files including file mask
            output (str): Absolute path to output file
            hflip (bool): Perform horizontal flip on input material.
            vflip (bool): Perform vertical flip on input material.
            rotate (bool): Perform 90° CCW rotation on input material.
            watermark (str): Path to watermark to apply to lower left corner.
            pixfmt (str): Pixel format to use for output. Default of yuv420p should usually fit the bill.

        Returns:
            (str): Prepared command string to render `input` to `output` using ffmpeg.
        """

        ### See unit tests in test/timelapse/test_timelapse_renderjob.py

        logger = logging.getLogger(__name__)

        ### Not all players can handle non-mpeg2 in VOB format
        if videocodec == "mpeg2video":
            containerformat = "vob"
        else:
            containerformat = "mp4"

        command = [
            ffmpeg,
            "-framerate",
            str(fps),
            "-i",
            '"{}"'.format(input),
            "-vcodec",
            videocodec,
            "-threads",
            str(threads),
            "-r",
            "25",
            "-y",
            "-b",
            str(bitrate),
            "-f",
            containerformat,
        ]

        filter_string = cls._create_filter_string(
            hflip=hflip, vflip=vflip, rotate=rotate, watermark=watermark
        )

        if filter_string is not None:
            logger.debug("Applying videofilter chain: {}".format(filter_string))
            command.extend(["-vf", sarge.shell_quote(filter_string)])

        # finalize command with output file
        logger.debug("Rendering movie to {}".format(output))
        command.append('"{}"'.format(output))

        return " ".join(command)

    @classmethod
    def _create_filter_string(
        cls, hflip=False, vflip=False, rotate=False, watermark=None, pixfmt="yuv420p"
    ):
        """
        Creates an ffmpeg filter string based on input parameters.

        Arguments:
            hflip (bool): Perform horizontal flip on input material.
            vflip (bool): Perform vertical flip on input material.
            rotate (bool): Perform 90° CCW rotation on input material.
            watermark (str): Path to watermark to apply to lower left corner.
            pixfmt (str): Pixel format to use, defaults to "yuv420p" which should usually fit the bill

        Returns:
            (str or None): filter string or None if no filters are required
        """

        ### See unit tests in test/timelapse/test_timelapse_renderjob.py

        # apply pixel format
        filters = ["format={}".format(pixfmt)]

        # flip video if configured
        if hflip:
            filters.append("hflip")
        if vflip:
            filters.append("vflip")
        if rotate:
            filters.append("transpose=2")

        # add watermark if configured
        watermark_filter = None
        if watermark is not None:
            watermark_filter = "movie={} [wm]; [{{input_name}}][wm] overlay=10:main_h-overlay_h-10".format(
                watermark
            )

        filter_string = None
        if len(filters) > 0:
            if watermark_filter is not None:
                filter_string = "[in] {} [postprocessed]; {} [out]".format(
                    ",".join(filters), watermark_filter.format(input_name="postprocessed")
                )
            else:
                filter_string = "[in] {} [out]".format(",".join(filters))
        elif watermark_filter is not None:
            filter_string = watermark_filter.format(input_name="in") + " [out]"

        return filter_string

    def _notify_callback(self, callback, *args, **kwargs):
        """Notifies registered callbacks of type `callback`."""
        name = "_on_{}".format(callback)
        method = getattr(self, name, None)
        if method is not None and callable(method):
            method(*args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2018 The OctoPrint Project - Released under terms of the AGPLv3 License"


import time

from octoprint.util import monotonic_time


def can_perform_update(target, check, online=True):
    return True


def perform_update(target, check, target_version, log_cb=None, online=True):
    duration = check.get("duration", 30)

    now = monotonic_time()
    end = now + duration
    while now < end:
        log_cb(["{}s left...".format(end - now)], prefix=">", stream="output")
        time.sleep(5)
        now = monotonic_time()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"


import logging
import re
import time

import sarge
from past.builtins import unicode

from octoprint.util.platform import CLOSE_FDS

from . import to_native_str, to_unicode
from .platform import get_os

try:
    import queue
except ImportError:  # pragma: no cover
    import Queue as queue

# These regexes are based on the colorama package
# Author: Jonathan Hartley
# License: BSD-3 (https://github.com/tartley/colorama/blob/master/LICENSE.txt)
# Website: https://github.com/tartley/colorama/
_ANSI_CSI_PATTERN = (
    b"\001?\033\\[(\\??(?:\\d|;)*)([a-zA-Z])\002?"  # Control Sequence Introducer
)
_ANSI_OSC_PATTERN = b"\001?\033\\]((?:.|;)*?)(\x07)\002?"  # Operating System Command
_ANSI_REGEX = re.compile(b"|".join([_ANSI_CSI_PATTERN, _ANSI_OSC_PATTERN]))


def clean_ansi(line):
    """
    Removes ANSI control codes from ``line``.

    Parameters:
        line (bytes or unicode): the line to process

    Returns:
        (bytes or unicode) The line without any ANSI control codes

    Example::

        >>> text = b"Some text with some \x1b[31mred words\x1b[39m in it"
        >>> clean_ansi(text) # doctest: +ALLOW_BYTES
        'Some text with some red words in it'
        >>> text = b"We \x1b[?25lhide the cursor here and then \x1b[?25hshow it again here"
        >>> clean_ansi(text) # doctest: +ALLOW_BYTES
        'We hide the cursor here and then show it again here'
    """
    if isinstance(line, unicode):
        return _ANSI_REGEX.sub(b"", line.encode("latin1")).decode("latin1")
    return _ANSI_REGEX.sub(b"", line)


class CommandlineError(Exception):
    """
    Raised by :py:func:`~octoprint.util.commandline.CommandLineCaller.checked_call` on non zero return codes

    Arguments:
        returncode (int): the return code of the command
        stdout (str): the stdout output produced by the command
        stderr (str): the stderr output produced by the command
    """

    def __init__(self, returncode, stdout, stderr):
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr


class CommandlineCaller(object):
    """
    The CommandlineCaller is a utility class that allows running command line commands while logging their stdout
    and stderr via configurable callback functions.

    Callbacks are expected to have a signature matching

    .. code-block:: python

       def callback(*lines):
           do_something_with_the_passed_lines()

    The class utilizes sarge underneath.

    Example:

    .. code-block:: python

       from octoprint.util.commandline import CommandLineCaller, CommandLineError

       def log(prefix, *lines):
           for line in lines:
               print("{} {}".format(prefix, line))

       def log_stdout(*lines):
           log(">>>", *lines)

       def log_stderr(*lines):
           log("!!!", *lines)

       def log_call(*lines)
           log("---", *lines)

       caller = CommandLineCaller()
       caller.on_log_call = log_call
       caller.on_log_stdout = log_stdout
       caller.on_log_stderr = log_stderr

       try:
           caller.checked_call(["some", "command", "with", "parameters"])
       except CommandLineError as err:
           print("Command returned {}".format(err.returncode))
       else:
           print("Command finished successfully")
    """

    def __init__(self):
        self._logger = logging.getLogger(__name__)

        self.on_log_call = lambda *args, **kwargs: None
        """Callback for the called command line"""

        self.on_log_stdout = lambda *args, **kwargs: None
        """Callback for stdout output"""

        self.on_log_stderr = lambda *args, **kwargs: None
        """Callback for stderr output"""

    def checked_call(self, command, **kwargs):
        """
        Calls a command and raises an error if it doesn't return with return code 0

        Args:
            command (list, tuple or str): command to call
            kwargs (dict): additional keyword arguments to pass to the sarge ``run`` call (note that ``_async``,
                           ``stdout`` and ``stderr`` will be overwritten)

        Returns:
            (tuple) a 3-tuple of return code, full stdout and full stderr output

        Raises:
            CommandlineError
        """
        returncode, stdout, stderr = self.call(command, **kwargs)

        if returncode != 0:
            raise CommandlineError(returncode, stdout, stderr)

        return returncode, stdout, stderr

    def call(self, command, **kwargs):
        """
        Calls a command

        Args:
            command (list, tuple or str): command to call
            kwargs (dict): additional keyword arguments to pass to the sarge ``run`` call (note that ``_async``,
                           ``stdout`` and ``stderr`` will be overwritten)

        Returns:
            (tuple) a 3-tuple of return code, full stdout and full stderr output
        """

        if isinstance(command, (list, tuple)):
            joined_command = " ".join(command)
        else:
            joined_command = command
        self._logger.debug("Calling: {}".format(joined_command))
        self.on_log_call(joined_command)

        # if we are running under windows, make sure there are no unicode strings in the env
        if get_os() == "windows" and "env" in kwargs:
            kwargs["env"] = dict(
                (to_native_str(k), to_native_str(v)) for k, v in kwargs["env"].items()
            )

        delimiter = kwargs.get("delimiter", b"\n")
        try:
            kwargs.pop("delimiter")
        except KeyError:
            pass

        buffer_size = kwargs.get("buffer_size", -1)
        try:
            kwargs.pop("buffer_size")
        except KeyError:
            pass

        kwargs.update(
            {
                "close_fds": CLOSE_FDS,
                "async_": True,
                "stdout": DelimiterCapture(delimiter=delimiter, buffer_size=buffer_size),
                "stderr": DelimiterCapture(delimiter=delimiter, buffer_size=buffer_size),
            }
        )

        p = sarge.run(command, **kwargs)
        while len(p.commands) == 0:
            # somewhat ugly... we can't use wait_events because
            # the events might not be all set if an exception
            # by sarge is triggered within the async process
            # thread
            time.sleep(0.01)

        # by now we should have a command, let's wait for its
        # process to have been prepared
        p.commands[0].process_ready.wait()

        if not p.commands[0].process:
            # the process might have been set to None in case of any exception
            self._logger.error(
                "Error while trying to run command {}".format(joined_command)
            )
            return None, [], []

        all_stdout = []
        all_stderr = []

        def process_lines(lines, logger):
            if not lines:
                return []
            processed = self._preprocess_lines(
                *map(lambda x: to_unicode(x, errors="replace"), lines)
            )
            logger(*processed)
            return list(processed)

        def process_stdout(lines):
            return process_lines(lines, self._log_stdout)

        def process_stderr(lines):
            return process_lines(lines, self._log_stderr)

        try:
            while p.returncode is None:
                all_stderr += process_stderr(p.stderr.readlines(timeout=0.5))
                all_stdout += process_stdout(p.stdout.readlines(timeout=0.5))
                p.commands[0].poll()

        finally:
            p.close()

        all_stderr += process_stderr(p.stderr.readlines())
        all_stdout += process_stdout(p.stdout.readlines())

        return p.returncode, all_stdout, all_stderr

    def _log_stdout(self, *lines):
        self.on_log_stdout(*lines)

    def _log_stderr(self, *lines):
        self.on_log_stderr(*lines)

    def _preprocess_lines(self, *lines):
        return lines


class DelimiterCapture(sarge.Capture):
    def __init__(self, delimiter=b"\n", *args, **kwargs):
        self._delimiter = delimiter
        sarge.Capture.__init__(self, *args, **kwargs)

    def readline(self, size=-1, block=True, timeout=None):
        if not self.streams_open():
            block = False
            timeout = None
        else:
            timeout = timeout or self.timeout
        if self.current is None:
            try:
                self.current = self.buffer.get(block, timeout)
            except queue.Empty:
                self.current = b""
        while self._delimiter not in self.current:
            try:
                self.current += self.buffer.get(block, timeout)
            except queue.Empty:
                break
        if self._delimiter not in self.current:
            result = self.current
            self.current = None
        else:
            i = self.current.index(self._delimiter)
            if 0 < size < i:
                i = size - 1
            result = self.current[: i + 1]
            self.current = self.current[i + 1 :]
        return result



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import octoprint.vendor.sockjs.tornado.transports.pollingbase

from .xhr import XhrPollingTransport, XhrSendHandler
from .jsonp import JSONPTransport, JSONPSendHandler
from .websocket import WebSocketTransport
from .xhrstreaming import XhrStreamingTransport
from .eventsource import EventSourceTransport
from .htmlfile import HtmlFileTransport
from .rawwebsocket import RawWebSocketTransport



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"


class FileDestinations(object):

    SDCARD = "sdcard"
    LOCAL = "local"



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"

import io
import logging
import os
import tarfile
import zipfile

try:
    from os import scandir
except ImportError:
    from scandir import scandir

from collections import defaultdict

from flask import jsonify, make_response, request
from flask_babel import Locale

from octoprint.access.permissions import Permissions
from octoprint.plugin import plugin_manager
from octoprint.server.api import api
from octoprint.server.util.flask import no_firstrun_access
from octoprint.settings import settings


@api.route("/languages", methods=["GET"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def getInstalledLanguagePacks():
    translation_folder = settings().getBaseFolder("translations", check_writable=False)
    if not os.path.exists(translation_folder):
        return jsonify(language_packs={"_core": []})

    core_packs = []
    plugin_packs = defaultdict(
        lambda: {"identifier": None, "display": None, "languages": []}
    )
    for entry in scandir(translation_folder):
        if not entry.is_dir():
            continue

        def load_meta(path, locale):
            meta = {}

            meta_path = os.path.join(path, "meta.yaml")
            if os.path.isfile(meta_path):
                import yaml

                try:
                    with io.open(meta_path, "rt", encoding="utf-8") as f:
                        meta = yaml.safe_load(f)
                except Exception:
                    logging.getLogger(__name__).exception("Could not load %s", meta_path)
                    pass
                else:
                    import datetime

                    if "last_update" in meta and isinstance(
                        meta["last_update"], datetime.datetime
                    ):
                        meta["last_update"] = (
                            meta["last_update"] - datetime.datetime(1970, 1, 1)
                        ).total_seconds()

            loc = Locale.parse(locale)
            meta["locale"] = locale
            meta["locale_display"] = loc.display_name
            meta["locale_english"] = loc.english_name
            return meta

        if entry.name == "_plugins":
            for plugin_entry in scandir(entry.path):
                if not plugin_entry.is_dir():
                    continue

                if plugin_entry.name not in plugin_manager().plugins:
                    continue

                plugin_info = plugin_manager().plugins[plugin_entry.name]

                plugin_packs[plugin_entry.name]["identifier"] = plugin_entry.name
                plugin_packs[plugin_entry.name]["display"] = plugin_info.name

                for language_entry in scandir(plugin_entry.path):
                    try:
                        plugin_packs[plugin_entry.name]["languages"].append(
                            load_meta(language_entry.path, language_entry.name)
                        )
                    except Exception:
                        logging.getLogger(__name__).exception(
                            "Error while parsing metadata for language pack {} from {} for plugin {}".format(
                                language_entry.name,
                                language_entry.path,
                                plugin_entry.name,
                            )
                        )
                        continue
        else:
            try:
                core_packs.append(load_meta(entry.path, entry.name))
            except ValueError:
                logging.getLogger(__name__).exception(
                    "Core language pack {} doesn't appear to actually be one".format(
                        entry.name
                    )
                )
            except Exception:
                logging.getLogger(__name__).exception(
                    "Error while parsing metadata for core language pack {} from {}".format(
                        entry.name, entry.path
                    )
                )

    result = {
        "_core": {"identifier": "_core", "display": "Core", "languages": core_packs}
    }
    result.update(plugin_packs)
    return jsonify(language_packs=result)


@api.route("/languages", methods=["POST"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def uploadLanguagePack():
    input_name = "file"
    input_upload_path = (
        input_name + "." + settings().get(["server", "uploads", "pathSuffix"])
    )
    input_upload_name = (
        input_name + "." + settings().get(["server", "uploads", "nameSuffix"])
    )
    if input_upload_path not in request.values or input_upload_name not in request.values:
        return make_response("No file included", 400)

    upload_name = request.values[input_upload_name]
    upload_path = request.values[input_upload_path]

    exts = list(
        filter(
            lambda x: upload_name.lower().endswith(x), (".zip", ".tar.gz", ".tgz", ".tar")
        )
    )
    if not len(exts):
        return make_response(
            "File doesn't have a valid extension for a language pack archive", 400
        )

    target_path = settings().getBaseFolder("translations")

    if tarfile.is_tarfile(upload_path):
        _unpack_uploaded_tarball(upload_path, target_path)
    elif zipfile.is_zipfile(upload_path):
        _unpack_uploaded_zipfile(upload_path, target_path)
    else:
        return make_response("Neither zip file nor tarball included", 400)

    return getInstalledLanguagePacks()


@api.route("/languages/<string:locale>/<string:pack>", methods=["DELETE"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def deleteInstalledLanguagePack(locale, pack):

    if pack == "_core":
        target_path = os.path.join(settings().getBaseFolder("translations"), locale)
    else:
        target_path = os.path.join(
            settings().getBaseFolder("translations"), "_plugins", pack, locale
        )

    if os.path.isdir(target_path):
        import shutil

        shutil.rmtree(target_path)

    return getInstalledLanguagePacks()


def _unpack_uploaded_zipfile(path, target):
    with zipfile.ZipFile(path, "r") as zip:
        # sanity check
        map(_validate_archive_name, zip.namelist())

        # unpack everything
        zip.extractall(target)


def _unpack_uploaded_tarball(path, target):
    with tarfile.open(path, "r") as tar:
        # sanity check
        map(_validate_archive_name, tar.getmembers())

        # unpack everything
        tar.extractall(target)


def _validate_archive_name(name):
    if name.startswith("/") or ".." in name:
        raise InvalidLanguagePack(
            "Provided language pack contains invalid name {name}".format(**locals())
        )


class InvalidLanguagePack(Exception):
    pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"


import io
import logging
import os
import site
import sys
import threading

import pkg_resources
import sarge

from octoprint.util import to_native_str, to_unicode
from octoprint.util.platform import CLOSE_FDS

from .commandline import CommandlineCaller, clean_ansi

_cache = {"version": {}, "setup": {}}
_cache_mutex = threading.RLock()


class UnknownPip(Exception):
    pass


class PipCaller(CommandlineCaller):
    process_dependency_links = pkg_resources.Requirement.parse("pip>=1.5")
    no_cache_dir = pkg_resources.Requirement.parse("pip>=1.6")
    disable_pip_version_check = pkg_resources.Requirement.parse("pip>=6.0")
    no_use_wheel = pkg_resources.Requirement.parse("pip==1.5.0")
    broken = pkg_resources.Requirement.parse("pip>=6.0.1,<=6.0.3")

    @classmethod
    def clean_install_command(cls, args, pip_version, virtual_env, use_user, force_user):
        logger = logging.getLogger(__name__)
        args = list(args)

        # strip --process-dependency-links for versions that don't support it
        if (
            pip_version not in cls.process_dependency_links
            and "--process-dependency-links" in args
        ):
            logger.debug(
                "Found --process-dependency-links flag, version {} doesn't need that yet though, removing.".format(
                    pip_version
                )
            )
            args.remove("--process-dependency-links")

        # strip --no-cache-dir for versions that don't support it
        if pip_version not in cls.no_cache_dir and "--no-cache-dir" in args:
            logger.debug(
                "Found --no-cache-dir flag, version {} doesn't support that yet though, removing.".format(
                    pip_version
                )
            )
            args.remove("--no-cache-dir")

        # strip --disable-pip-version-check for versions that don't support it
        if (
            pip_version not in cls.disable_pip_version_check
            and "--disable-pip-version-check" in args
        ):
            logger.debug(
                "Found --disable-pip-version-check flag, version {} doesn't support that yet though, removing.".format(
                    pip_version
                )
            )
            args.remove("--disable-pip-version-check")

        # add --no-use-wheel for versions that otherwise break
        if pip_version in cls.no_use_wheel and "--no-use-wheel" not in args:
            logger.debug(
                "Version {} needs --no-use-wheel to properly work.".format(pip_version)
            )
            args.append("--no-use-wheel")

        # remove --user if it's present and a virtual env is detected
        if "--user" in args:
            if virtual_env or not site.ENABLE_USER_SITE:
                logger.debug("Virtual environment detected, removing --user flag.")
                args.remove("--user")
        # otherwise add it if necessary
        elif not virtual_env and site.ENABLE_USER_SITE and (use_user or force_user):
            logger.debug("pip needs --user flag for installations.")
            args.append("--user")

        return args

    def __init__(
        self, configured=None, ignore_cache=False, force_sudo=False, force_user=False
    ):
        CommandlineCaller.__init__(self)
        self._logger = logging.getLogger(__name__)

        self.configured = configured
        self.refresh = False
        self.ignore_cache = ignore_cache

        self.force_sudo = force_sudo
        self.force_user = force_user

        self._command = None
        self._version = None
        self._version_string = None
        self._use_sudo = False
        self._use_user = False
        self._virtual_env = False
        self._install_dir = None

        self.trigger_refresh()

        self.on_log_call = lambda *args, **kwargs: None
        self.on_log_stdout = lambda *args, **kwargs: None
        self.on_log_stderr = lambda *args, **kwargs: None

    def _reset(self):
        self._command = None
        self._version = None
        self._version_string = None
        self._use_sudo = False
        self._use_user = False
        self._install_dir = None

    def __le__(self, other):
        return self.version is not None and self.version <= other

    def __lt__(self, other):
        return self.version is not None and self.version < other

    def __ge__(self, other):
        return self.version is not None and self.version >= other

    def __gt__(self, other):
        return self.version is not None and self.version > other

    @property
    def command(self):
        return self._command

    @property
    def version(self):
        return self._version

    @property
    def version_string(self):
        return self._version_string

    @property
    def install_dir(self):
        return self._install_dir

    @property
    def use_sudo(self):
        return self._use_sudo

    @property
    def use_user(self):
        return self._use_user

    @property
    def virtual_env(self):
        return self._virtual_env

    @property
    def available(self):
        return self._command is not None

    def trigger_refresh(self):
        self._reset()
        try:
            self._setup_pip()
        except Exception:
            self._logger.exception("Error while discovering pip command")
            self._command = None
            self._version = None
        self.refresh = False

    def execute(self, *args, **kwargs):
        if self.refresh:
            self.trigger_refresh()

        if self._command is None:
            raise UnknownPip()

        arg_list = list(args)

        if "install" in arg_list:
            arg_list = self.clean_install_command(
                arg_list, self.version, self._virtual_env, self.use_user, self.force_user
            )

        # add args to command
        if isinstance(self._command, list):
            command = self._command + list(arg_list)
        else:
            command = [self._command] + list(arg_list)

        # add sudo if necessary
        if self._use_sudo or self.force_sudo:
            command = ["sudo"] + command

        return self.call(command, **kwargs)

    def _setup_pip(self):
        pip_command, pip_sudo = self._get_pip_command()
        if pip_command is None:
            return

        # Determine the pip version

        self._logger.debug("Going to figure out pip's version")

        pip_version, version_segment = self._get_pip_version(pip_command)
        if pip_version is None:
            return

        if pip_version in self.__class__.broken:
            self._logger.error(
                "This version of pip is known to have bugs that make it incompatible with how it needs "
                "to be used by OctoPrint. Please upgrade your pip version."
            )
            return

        # Now figure out if pip belongs to a virtual environment and if the
        # default installation directory is writable.
        #
        # The idea is the following: If OctoPrint is installed globally,
        # the site-packages folder is probably not writable by our user.
        # However, the user site-packages folder as usable by providing the
        # --user parameter during install is. This we may not use though if
        # the provided pip belongs to a virtual env (since that hiccups hard).
        #
        # So we figure out the installation directory, check if it's writable
        # and if not if pip belongs to a virtual environment. Only if the
        # installation directory is NOT writable by us but we also don't run
        # in a virtual environment may we proceed with the --user parameter.

        ok, pip_user, pip_virtual_env, pip_install_dir = self._check_pip_setup(
            pip_command
        )
        if not ok:
            if pip_install_dir:
                self._logger.error(
                    "Cannot use this pip install, can't write to the install dir and also can't use "
                    "--user for installing. Check your setup and the permissions on {}.".format(
                        pip_install_dir
                    )
                )
            else:
                self._logger.error(
                    "Cannot use this pip install, something's wrong with the python environment. "
                    "Check the lines before."
                )
            return

        self._command = pip_command
        self._version = pip_version
        self._version_string = version_segment
        self._use_sudo = pip_sudo
        self._use_user = pip_user
        self._virtual_env = pip_virtual_env
        self._install_dir = pip_install_dir

    def _get_pip_command(self):
        pip_command = self.configured

        if pip_command is not None and pip_command.startswith("sudo "):
            pip_command = pip_command[len("sudo ") :]
            pip_sudo = True
        else:
            pip_sudo = False

        if pip_command is None:
            pip_command = self.autodetect_pip()

        return pip_command, pip_sudo

    @classmethod
    def autodetect_pip(cls):
        commands = [
            [sys.executable, "-m", "pip"],
            [
                os.path.join(
                    os.path.dirname(sys.executable),
                    "pip.exe" if sys.platform == "win32" else "pip",
                )
            ],
            # this should be our last resort since it might fail thanks to using pip programmatically like
            # that is not officially supported or sanctioned by the pip developers
            [
                sys.executable,
                "-c",
                "import sys; sys.argv = ['pip'] + sys.argv[1:]; import pip; pip.main()",
            ],
        ]

        for command in commands:
            p = sarge.run(
                command + ["--version"],
                close_fds=CLOSE_FDS,
                stdout=sarge.Capture(),
                stderr=sarge.Capture(),
            )
            if p.returncode == 0:
                logging.getLogger(__name__).info(
                    'Using "{}" as command to invoke pip'.format(" ".join(command))
                )
                return command

        return None

    @classmethod
    def to_sarge_command(cls, pip_command, *args):
        if isinstance(pip_command, list):
            sarge_command = pip_command
        else:
            sarge_command = [pip_command]
        return sarge_command + list(args)

    def _get_pip_version(self, pip_command):
        # Debugging this with PyCharm/IntelliJ with Python plugin and no output is being
        # generated? PyCharm bug. Disable "Attach to subprocess automatically when debugging"
        # in IDE Settings or patch pydevd.py
        # -> https://youtrack.jetbrains.com/issue/PY-18365#comment=27-1290453

        pip_command_str = pip_command
        if isinstance(pip_command_str, list):
            pip_command_str = " ".join(pip_command_str)

        with _cache_mutex:
            if not self.ignore_cache and pip_command_str in _cache["version"]:
                self._logger.debug(
                    "Using cached pip version information for {}".format(pip_command_str)
                )
                return _cache["version"][pip_command_str]

            sarge_command = self.to_sarge_command(pip_command, "--version")
            p = sarge.run(
                sarge_command,
                close_fds=CLOSE_FDS,
                stdout=sarge.Capture(),
                stderr=sarge.Capture(),
            )

            if p.returncode != 0:
                self._logger.warning(
                    "Error while trying to run pip --version: {}".format(p.stderr.text)
                )
                return None, None

            output = PipCaller._preprocess(p.stdout.text)
            # output should look something like this:
            #
            #     pip <version> from <path> (<python version>)
            #
            # we'll just split on whitespace and then try to use the second entry

            if not output.startswith("pip"):
                self._logger.warning(
                    "pip command returned unparseable output, can't determine version: {}".format(
                        output
                    )
                )

            split_output = list(map(lambda x: x.strip(), output.split()))
            if len(split_output) < 2:
                self._logger.warning(
                    "pip command returned unparseable output, can't determine version: {}".format(
                        output
                    )
                )

            version_segment = split_output[1]

            try:
                pip_version = pkg_resources.parse_version(version_segment)
            except Exception:
                self._logger.exception(
                    "Error while trying to parse version string from pip command"
                )
                return None, None

            self._logger.info("Version of pip is {}".format(version_segment))

            result = pip_version, version_segment
            _cache["version"][pip_command_str] = result
            return result

    def _check_pip_setup(self, pip_command):
        pip_command_str = pip_command
        if isinstance(pip_command_str, list):
            pip_command_str = " ".join(pip_command_str)

        with _cache_mutex:
            if not self.ignore_cache and pip_command_str in _cache["setup"]:
                self._logger.debug(
                    "Using cached pip setup information for {}".format(pip_command_str)
                )
                return _cache["setup"][pip_command_str]

            # This is horribly ugly and I'm sorry...
            #
            # While we can figure out the install directory, if that's writable and if a virtual environment
            # is active for pip that belongs to our sys.executable python instance by just checking some
            # variables, we can't for stuff like third party software we allow to update via the software
            # update plugin.
            #
            # What we do instead for these situations is try to install (and of course uninstall) the
            # testballoon dummy package, which collects that information for us. For pip <= 7 we could
            # have the testballoon provide us with the info needed through stdout, if pip was called
            # with --verbose anything printed to stdout within setup.py would be output. Pip 8 managed
            # to break this mechanism. Any (!) output within setup.py appears to be suppressed now, and
            # no combination of --log and multiple --verbose or -v arguments could get it to bring the
            # output back.
            #
            # So here's what we do now instead. Our sarge call sets an environment variable
            # "TESTBALLOON_OUTPUT" that points to a temporary file. If the testballoon sees that
            # variable set, it opens the file and writes to it the output it so far printed on stdout.
            # We then open the file and read in the data that way.
            #
            # Yeah, I'm not happy with that either. But as long as there's no way to otherwise figure
            # out for a generic pip command whether OctoPrint can even install anything with that
            # and if so how, well, that's how we'll have to do things.

            import os

            testballoon = os.path.join(
                os.path.realpath(os.path.dirname(__file__)), "piptestballoon"
            )

            from octoprint.util import temppath

            with temppath() as testballoon_output_file:
                sarge_command = self.to_sarge_command(pip_command, "install", ".")
                try:
                    # our testballoon is no real package, so this command will fail - that's ok though,
                    # we only need the output produced within the pip environment
                    sarge.run(
                        sarge_command,
                        close_fds=CLOSE_FDS,
                        stdout=sarge.Capture(),
                        stderr=sarge.Capture(),
                        cwd=testballoon,
                        env={
                            to_native_str("TESTBALLOON_OUTPUT"): to_native_str(
                                testballoon_output_file
                            )
                        },
                    )
                except Exception:
                    self._logger.exception(
                        "Error while trying to install testballoon to figure out pip setup"
                    )
                    return False, False, False, None

                data = {}
                with io.open(testballoon_output_file, "rt", encoding="utf-8") as f:
                    for line in f:
                        key, value = line.split("=", 2)
                        data[key] = value

            install_dir_str = data.get("PIP_INSTALL_DIR", None)
            virtual_env_str = data.get("PIP_VIRTUAL_ENV", None)
            writable_str = data.get("PIP_WRITABLE", None)

            if (
                install_dir_str is not None
                and virtual_env_str is not None
                and writable_str is not None
            ):
                install_dir = install_dir_str.strip()
                virtual_env = virtual_env_str.strip() == "True"
                writable = writable_str.strip() == "True"

                can_use_user_flag = not virtual_env and site.ENABLE_USER_SITE

                ok = writable or can_use_user_flag
                user_flag = not writable and can_use_user_flag

                self._logger.info(
                    "pip installs to {} (writable -> {}), --user flag needed -> {}, "
                    "virtual env -> {}".format(
                        install_dir,
                        "yes" if writable else "no",
                        "yes" if user_flag else "no",
                        "yes" if virtual_env else "no",
                    )
                )
                self._logger.info("==> pip ok -> {}".format("yes" if ok else "NO!"))

                # ok, enable user flag, virtual env yes/no, installation dir
                result = ok, user_flag, virtual_env, install_dir
                _cache["setup"][pip_command_str] = result
                return result
            else:
                self._logger.error(
                    "Could not detect desired output from testballoon install, got this instead: {!r}".format(
                        data
                    )
                )
                return False, False, False, None

    def _preprocess_lines(self, *lines):
        return list(map(self._preprocess, lines))

    @staticmethod
    def _preprocess(text):
        """
        Strips ANSI and VT100 cursor control characters from line and makes sure it's a unicode.

        Parameters:
            text (str or unicode): The text to process

        Returns:
            (unicode) The processed text as a unicode, stripped of ANSI and VT100 cursor show/hide codes

        Example::

            >>> text = b'some text with some\x1b[?25h ANSI codes for \x1b[31mred words\x1b[39m and\x1b[?25l also some cursor control codes'
            >>> PipCaller._preprocess(text) # doctest: +ALLOW_UNICODE
            'some text with some ANSI codes for red words and also some cursor control codes'
        """
        return to_unicode(clean_ansi(text))


class LocalPipCaller(PipCaller):
    """
    The LocalPipCaller always uses the pip instance associated with
    sys.executable.
    """

    def _get_pip_command(self):
        return self.autodetect_pip(), False

    def _check_pip_setup(self, pip_command):
        import os
        import sys
        from distutils.sysconfig import get_python_lib

        virtual_env = hasattr(sys, "real_prefix") or (
            hasattr(sys, "base_prefix") and sys.base_prefix != sys.prefix
        )
        install_dir = get_python_lib()
        writable = os.access(install_dir, os.W_OK)

        can_use_user_flag = not virtual_env and site.ENABLE_USER_SITE
        user_flag = not writable and can_use_user_flag

        ok = writable or can_use_user_flag

        self._logger.info(
            "pip installs to {} (writable -> {}), --user flag needed -> {}, "
            "virtual env -> {}".format(
                install_dir,
                "yes" if writable else "no",
                "yes" if user_flag else "no",
                "yes" if virtual_env else "no",
            )
        )
        self._logger.info("==> pip ok -> {}".format("yes" if ok else "NO!"))

        return ok, user_flag, virtual_env, install_dir


def create_pip_caller(command=None, **kwargs):
    if command is None:
        return LocalPipCaller(**kwargs)
    else:
        return PipCaller(configured=command, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import io
import logging as log
import os
import sys

from ._version import get_versions

# ~~ version


versions = get_versions()

__version__ = versions["version"]
__branch__ = versions.get("branch", None)
__display_version__ = __version__
__revision__ = versions.get("full-revisionid", versions.get("full", None))

del versions
del get_versions

# figure out current umask - sadly only doable by setting a new one and resetting it, no query method
UMASK = os.umask(0)
os.umask(UMASK)

urllib3_ssl = True
"""Whether requests/urllib3 and urllib3 (if installed) should be able to establish
   a sound SSL environment or not."""

version_info = sys.version_info
if version_info.major == 2 and version_info.minor <= 7 and version_info.micro < 9:
    try:
        # make sure our requests version of urllib3 is properly patched (if possible)
        import requests.packages.urllib3.contrib.pyopenssl

        requests.packages.urllib3.contrib.pyopenssl.inject_into_urllib3()
    except ImportError:
        urllib3_ssl = False

    try:
        import urllib3

        # only proceed if urllib3 is even installed on its own
        try:
            # urllib3 is there, let's patch that too
            import urllib3.contrib.pyopenssl

            urllib3.contrib.pyopenssl.inject_into_urllib3()
        except ImportError:
            urllib3_ssl = False
    except ImportError:
        pass

elif version_info.major == 3 and version_info.minor >= 8 and sys.platform == "win32":
    # Python 3.8 makes proactor event loop the default on Windows, Tornado doesn't like that
    #
    # see https://github.com/tornadoweb/tornado/issues/2608
    import asyncio

    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

del version_info

# ~~ custom exceptions


class FatalStartupError(Exception):
    def __init__(self, message, cause=None):
        self.cause = cause
        Exception.__init__(self, message)

    def __str__(self):
        result = Exception.__str__(self)
        if self.cause:
            return "{}: {}".format(result, str(self.cause))
        else:
            return result


# ~~ init methods to bring up platform


def init_platform(
    basedir,
    configfile,
    use_logging_file=True,
    logging_file=None,
    logging_config=None,
    debug=False,
    verbosity=0,
    uncaught_logger=None,
    uncaught_handler=None,
    safe_mode=False,
    ignore_blacklist=False,
    after_preinit_logging=None,
    after_settings_init=None,
    after_logging=None,
    after_safe_mode=None,
    after_settings_valid=None,
    after_event_manager=None,
    after_connectivity_checker=None,
    after_plugin_manager=None,
    after_environment_detector=None,
):
    kwargs = {}

    logger, recorder = preinit_logging(
        debug, verbosity, uncaught_logger, uncaught_handler
    )
    kwargs["logger"] = logger
    kwargs["recorder"] = recorder

    if callable(after_preinit_logging):
        after_preinit_logging(**kwargs)

    try:
        settings = init_settings(basedir, configfile)
    except Exception as ex:
        raise FatalStartupError("Could not initialize settings manager", cause=ex)
    kwargs["settings"] = settings
    if callable(after_settings_init):
        after_settings_init(**kwargs)

    try:
        logger = init_logging(
            settings,
            use_logging_file=use_logging_file,
            logging_file=logging_file,
            default_config=logging_config,
            debug=debug,
            verbosity=verbosity,
            uncaught_logger=uncaught_logger,
            uncaught_handler=uncaught_handler,
        )
    except Exception as ex:
        raise FatalStartupError("Could not initialize logging", cause=ex)

    kwargs["logger"] = logger
    if callable(after_logging):
        after_logging(**kwargs)

    settings_start_once_in_safemode = (
        "settings" if settings.getBoolean(["server", "startOnceInSafeMode"]) else None
    )
    settings_incomplete_startup_safemode = (
        "incomplete_startup"
        if settings.getBoolean(["server", "incompleteStartup"])
        and not settings.getBoolean(["server", "ignoreIncompleteStartup"])
        else None
    )
    safe_mode = (
        safe_mode
        or settings_start_once_in_safemode
        or settings_incomplete_startup_safemode
    )
    kwargs["safe_mode"] = safe_mode
    if callable(after_safe_mode):
        after_safe_mode(**kwargs)

    # now before we continue, let's make sure *all* our folders are sane
    try:
        settings.sanity_check_folders()
    except Exception as ex:
        raise FatalStartupError("Configured folders didn't pass sanity check", cause=ex)
    if callable(after_settings_valid):
        after_settings_valid(**kwargs)

    try:
        event_manager = init_event_manager(settings)
    except Exception as ex:
        raise FatalStartupError("Could not initialize event manager", cause=ex)

    kwargs["event_manager"] = event_manager
    if callable(after_event_manager):
        after_event_manager(**kwargs)

    try:
        connectivity_checker = init_connectivity_checker(settings, event_manager)
    except Exception as ex:
        raise FatalStartupError("Could not initialize connectivity checker", cause=ex)

    kwargs["connectivity_checker"] = connectivity_checker
    if callable(after_connectivity_checker):
        after_connectivity_checker(**kwargs)

    try:
        plugin_manager = init_pluginsystem(
            settings,
            safe_mode=safe_mode,
            ignore_blacklist=ignore_blacklist,
            connectivity_checker=connectivity_checker,
        )
    except Exception as ex:
        raise FatalStartupError("Could not initialize plugin manager", cause=ex)

    kwargs["plugin_manager"] = plugin_manager
    if callable(after_plugin_manager):
        after_plugin_manager(**kwargs)

    try:
        environment_detector = init_environment_detector(plugin_manager)
    except Exception as ex:
        raise FatalStartupError("Could not initialize environment detector", cause=ex)

    kwargs["environment_detector"] = environment_detector
    if callable(after_environment_detector):
        after_environment_detector(**kwargs)

    return (
        settings,
        logger,
        safe_mode,
        event_manager,
        connectivity_checker,
        plugin_manager,
        environment_detector,
    )


def init_settings(basedir, configfile):
    """Inits the settings instance based on basedir and configfile to use."""

    from octoprint.settings import InvalidSettings, settings

    try:
        return settings(init=True, basedir=basedir, configfile=configfile)
    except InvalidSettings as e:
        raise FatalStartupError(str(e))


def preinit_logging(
    debug=False, verbosity=0, uncaught_logger=None, uncaught_handler=None
):
    config = {
        "version": 1,
        "formatters": {
            "simple": {"format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s"}
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": "DEBUG",
                "formatter": "simple",
                "stream": "ext://sys.stdout",
            }
        },
        "loggers": {
            "octoprint": {"level": "DEBUG" if debug else "INFO"},
            "octoprint.util": {"level": "INFO"},
        },
        "root": {"level": "WARN", "handlers": ["console"]},
    }

    logger = set_logging_config(
        config, debug, verbosity, uncaught_logger, uncaught_handler
    )

    from octoprint.logging.handlers import RecordingLogHandler

    recorder = RecordingLogHandler(level=log.DEBUG)
    log.getLogger().addHandler(recorder)

    return logger, recorder


def init_logging(
    settings,
    use_logging_file=True,
    logging_file=None,
    default_config=None,
    debug=False,
    verbosity=0,
    uncaught_logger=None,
    uncaught_handler=None,
):
    """Sets up logging."""

    import os

    from octoprint.util import dict_merge

    # default logging configuration
    if default_config is None:
        default_config = {
            "version": 1,
            "formatters": {
                "simple": {
                    "format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
                },
                "serial": {"format": "%(asctime)s - %(message)s"},
                "timings": {"format": "%(asctime)s - %(message)s"},
                "timingscsv": {"format": "%(asctime)s;%(func)s;%(timing)f"},
            },
            "handlers": {
                "console": {
                    "class": "octoprint.logging.handlers.OctoPrintStreamHandler",
                    "level": "DEBUG",
                    "formatter": "simple",
                    "stream": "ext://sys.stdout",
                },
                "file": {
                    "class": "octoprint.logging.handlers.OctoPrintLogHandler",
                    "level": "DEBUG",
                    "formatter": "simple",
                    "when": "D",
                    "backupCount": 6,
                    "filename": os.path.join(
                        settings.getBaseFolder("logs"), "octoprint.log"
                    ),
                },
                "serialFile": {
                    "class": "octoprint.logging.handlers.SerialLogHandler",
                    "level": "DEBUG",
                    "formatter": "serial",
                    "backupCount": 3,
                    "filename": os.path.join(
                        settings.getBaseFolder("logs"), "serial.log"
                    ),
                    "delay": True,
                },
                "pluginTimingsFile": {
                    "class": "octoprint.logging.handlers.PluginTimingsLogHandler",
                    "level": "DEBUG",
                    "formatter": "timings",
                    "backupCount": 3,
                    "filename": os.path.join(
                        settings.getBaseFolder("logs"), "plugintimings.log"
                    ),
                    "delay": True,
                },
                "pluginTimingsCsvFile": {
                    "class": "octoprint.logging.handlers.PluginTimingsLogHandler",
                    "level": "DEBUG",
                    "formatter": "timingscsv",
                    "backupCount": 3,
                    "filename": os.path.join(
                        settings.getBaseFolder("logs"), "plugintimings.csv"
                    ),
                    "delay": True,
                },
            },
            "loggers": {
                "SERIAL": {
                    "level": "INFO",
                    "handlers": ["serialFile"],
                    "propagate": False,
                },
                "PLUGIN_TIMINGS": {
                    "level": "INFO",
                    "handlers": ["pluginTimingsFile", "pluginTimingsCsvFile"],
                    "propagate": False,
                },
                "PLUGIN_TIMINGS.octoprint.plugin": {"level": "INFO"},
                "octoprint": {"level": "INFO"},
                "octoprint.util": {"level": "INFO"},
                "octoprint.plugins": {"level": "INFO"},
            },
            "root": {"level": "WARN", "handlers": ["console", "file"]},
        }

    if debug or verbosity > 0:
        default_config["loggers"]["octoprint"]["level"] = "DEBUG"
        default_config["root"]["level"] = "INFO"
    if verbosity > 1:
        default_config["loggers"]["octoprint.plugins"]["level"] = "DEBUG"
    if verbosity > 2:
        default_config["root"]["level"] = "DEBUG"

    config = default_config
    if use_logging_file:
        # further logging configuration from file...
        if logging_file is None:
            logging_file = os.path.join(settings.getBaseFolder("base"), "logging.yaml")

        config_from_file = {}
        if os.path.exists(logging_file) and os.path.isfile(logging_file):
            import yaml

            with io.open(logging_file, "rt", encoding="utf-8") as f:
                config_from_file = yaml.safe_load(f)

        # we merge that with the default config
        if config_from_file is not None and isinstance(config_from_file, dict):
            config = dict_merge(default_config, config_from_file)

    # configure logging globally
    return set_logging_config(config, debug, verbosity, uncaught_logger, uncaught_handler)


def octoprint_plugin_inject_factory(settings, components):
    import octoprint.plugin

    def f(name, implementation):
        """Factory for injections for all OctoPrintPlugins"""
        if not isinstance(implementation, octoprint.plugin.OctoPrintPlugin):
            return None

        components_copy = dict(components)
        if "printer" in components:
            import functools

            import wrapt

            def tagwrap(f):
                @functools.wraps(f)
                def wrapper(*args, **kwargs):
                    tags = kwargs.get("tags", set()) | {
                        "source:plugin",
                        "plugin:{}".format(name),
                    }
                    kwargs["tags"] = tags
                    return f(*args, **kwargs)

                wrapper.__tagwrapped__ = True
                return wrapper

            class TaggedFuncsPrinter(wrapt.ObjectProxy):
                def __getattribute__(self, attr):
                    __wrapped__ = super(TaggedFuncsPrinter, self).__getattribute__(
                        "__wrapped__"
                    )
                    if attr == "__wrapped__":
                        return __wrapped__

                    item = getattr(__wrapped__, attr)
                    if (
                        callable(item)
                        and (
                            "tags" in item.__code__.co_varnames
                            or "kwargs" in item.__code__.co_varnames
                        )
                        and not getattr(item, "__tagwrapped__", False)
                    ):
                        return tagwrap(item)
                    else:
                        return item

            components_copy["printer"] = TaggedFuncsPrinter(components["printer"])

        props = {}
        props.update(components_copy)
        props.update({"data_folder": os.path.join(settings.getBaseFolder("data"), name)})
        return props

    return f


def settings_plugin_inject_factory(settings):
    import octoprint.plugin

    def f(name, implementation):
        """Factory for additional injections/initializations depending on plugin type"""
        if not isinstance(implementation, octoprint.plugin.SettingsPlugin):
            return

        default_settings_overlay = {"plugins": {}}
        default_settings_overlay["plugins"][name] = implementation.get_settings_defaults()
        settings.add_overlay(default_settings_overlay, at_end=True)

        plugin_settings = octoprint.plugin.plugin_settings_for_settings_plugin(
            name, implementation
        )
        if plugin_settings is None:
            return

        return {"settings": plugin_settings}

    return f


def init_settings_plugin_config_migration_and_cleanup(plugin_manager):
    import logging

    import octoprint.plugin

    def settings_plugin_config_migration_and_cleanup(identifier, implementation):
        """Take care of migrating and cleaning up any old settings"""

        if not isinstance(implementation, octoprint.plugin.SettingsPlugin):
            return

        settings_version = implementation.get_settings_version()
        settings_migrator = implementation.on_settings_migrate

        if settings_version is not None and settings_migrator is not None:
            stored_version = implementation._settings.get_int(
                [octoprint.plugin.SettingsPlugin.config_version_key]
            )
            if stored_version is None or stored_version < settings_version:
                settings_migrator(settings_version, stored_version)
                implementation._settings.set_int(
                    [octoprint.plugin.SettingsPlugin.config_version_key],
                    settings_version,
                    force=True,
                )

        implementation.on_settings_cleanup()
        implementation._settings.save()

        implementation.on_settings_initialized()

    settingsPlugins = plugin_manager.get_implementations(octoprint.plugin.SettingsPlugin)
    for implementation in settingsPlugins:
        try:
            settings_plugin_config_migration_and_cleanup(
                implementation._identifier, implementation
            )
        except Exception:
            logging.getLogger(__name__).exception(
                "Error while trying to migrate settings for "
                "plugin {}, ignoring it".format(implementation._identifier),
                extra={"plugin": implementation._identifier},
            )

    plugin_manager.implementation_post_inits = [
        settings_plugin_config_migration_and_cleanup
    ]


def init_custom_events(plugin_manager):
    import logging

    import octoprint.events

    logger = logging.getLogger(__name__)

    custom_events_hooks = plugin_manager.get_hooks(
        "octoprint.events.register_custom_events"
    )
    for name, hook in custom_events_hooks.items():
        try:
            result = hook()
            if isinstance(result, (list, tuple)):
                for event in result:
                    constant, value = octoprint.events.Events.register_event(
                        event, prefix="plugin_{}_".format(name)
                    )
                    logger.debug(
                        'Registered event {} of plugin {} as Events.{} = "{}"'.format(
                            event, name, constant, value
                        )
                    )
        except Exception:
            logger.exception(
                "Error while retrieving custom event list from plugin {}".format(name),
                extra={"plugin": name},
            )


def set_logging_config(config, debug, verbosity, uncaught_logger, uncaught_handler):
    # configure logging globally
    import logging.config as logconfig

    logconfig.dictConfig(config)

    # make sure we log any warnings
    log.captureWarnings(True)

    import warnings

    categories = (DeprecationWarning, PendingDeprecationWarning)
    if verbosity > 2:
        warnings.simplefilter("always")
    elif debug or verbosity > 0:
        for category in categories:
            warnings.simplefilter("always", category=category)

    # make sure we also log any uncaught exceptions
    if uncaught_logger is None:
        logger = log.getLogger(__name__)
    else:
        logger = log.getLogger(uncaught_logger)

    if uncaught_handler is None:

        def exception_logger(exc_type, exc_value, exc_tb):
            logger.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_tb))

        uncaught_handler = exception_logger
    sys.excepthook = uncaught_handler

    return logger


def init_pluginsystem(
    settings, safe_mode=False, ignore_blacklist=True, connectivity_checker=None
):
    """Initializes the plugin manager based on the settings."""

    import os

    # we need this so that octoprint.plugins is in sys.modules and no warnings are caused when loading bundled plugins
    import octoprint.plugins  # noqa: F401

    logger = log.getLogger(__name__ + ".startup")

    plugin_folders = [
        (
            os.path.abspath(
                os.path.join(os.path.dirname(os.path.realpath(__file__)), "plugins")
            ),
            "octoprint.plugins",
            True,
        ),
        settings.getBaseFolder("plugins"),
    ]
    plugin_entry_points = ["octoprint.plugin"]
    plugin_disabled_list = settings.get(["plugins", "_disabled"])

    plugin_blacklist = []
    if not ignore_blacklist and settings.getBoolean(
        ["server", "pluginBlacklist", "enabled"]
    ):
        plugin_blacklist = get_plugin_blacklist(
            settings, connectivity_checker=connectivity_checker
        )

    plugin_validators = []
    if safe_mode:

        def validator(phase, plugin_info):
            if phase in ("before_import", "before_load", "before_enable"):
                plugin_info.safe_mode_victim = not plugin_info.bundled
                if not plugin_info.bundled:
                    return False
            return True

        plugin_validators.append(validator)

    compatibility_ignored_list = settings.get(["plugins", "_forcedCompatible"])

    from octoprint.plugin import plugin_manager

    pm = plugin_manager(
        init=True,
        plugin_folders=plugin_folders,
        plugin_entry_points=plugin_entry_points,
        plugin_disabled_list=plugin_disabled_list,
        plugin_blacklist=plugin_blacklist,
        plugin_validators=plugin_validators,
        compatibility_ignored_list=compatibility_ignored_list,
    )

    settings_overlays = {}
    disabled_from_overlays = {}

    def handle_plugin_loaded(name, plugin):
        if plugin.instance and hasattr(plugin.instance, "__plugin_settings_overlay__"):
            plugin.needs_restart = True

            # plugin has a settings overlay, inject it
            overlay_definition = plugin.instance.__plugin_settings_overlay__
            if isinstance(overlay_definition, (tuple, list)):
                overlay_definition, order = overlay_definition
            else:
                order = None

            overlay = settings.load_overlay(overlay_definition)

            if "plugins" in overlay and "_disabled" in overlay["plugins"]:
                disabled_plugins = overlay["plugins"]["_disabled"]
                del overlay["plugins"]["_disabled"]
                disabled_from_overlays[name] = (disabled_plugins, order)

            settings_overlays[name] = overlay
            logger.debug("Found settings overlay on plugin {}".format(name))

    def handle_plugins_loaded(
        startup=False, initialize_implementations=True, force_reload=None
    ):
        if not startup:
            return

        from octoprint.util import sv

        sorted_disabled_from_overlays = sorted(
            [(key, value[0], value[1]) for key, value in disabled_from_overlays.items()],
            key=lambda x: (x[2] is None, sv(x[2]), sv(x[0])),
        )

        disabled_list = pm.plugin_disabled_list
        already_processed = []
        for name, addons, _ in sorted_disabled_from_overlays:
            if name not in disabled_list and not name.endswith("disabled"):
                for addon in addons:
                    if addon in disabled_list:
                        continue

                    if addon in already_processed:
                        logger.info(
                            "Plugin {} wants to disable plugin {}, but that was already processed".format(
                                name, addon
                            )
                        )

                    if addon not in already_processed and addon not in disabled_list:
                        disabled_list.append(addon)
                        logger.info(
                            "Disabling plugin {} as defined by plugin {}".format(
                                addon, name
                            )
                        )
                already_processed.append(name)

    def handle_plugin_enabled(name, plugin):
        if name in settings_overlays:
            settings.add_overlay(settings_overlays[name])
            logger.info("Added settings overlay from plugin {}".format(name))

    pm.on_plugin_loaded = handle_plugin_loaded
    pm.on_plugins_loaded = handle_plugins_loaded
    pm.on_plugin_enabled = handle_plugin_enabled
    pm.reload_plugins(startup=True, initialize_implementations=False)
    return pm


def get_plugin_blacklist(settings, connectivity_checker=None):
    import os
    import time

    import requests
    import yaml

    from octoprint.util import bom_aware_open
    from octoprint.util.version import is_octoprint_compatible

    logger = log.getLogger(__name__ + ".startup")

    if connectivity_checker is not None and not connectivity_checker.online:
        logger.info("We don't appear to be online, not fetching plugin blacklist")
        return []

    def format_blacklist(entries):
        format_entry = (
            lambda x: "{} ({})".format(x[0], x[1])
            if isinstance(x, (list, tuple)) and len(x) == 2
            else "{} (any)".format(x)
        )
        return ", ".join(map(format_entry, entries))

    def process_blacklist(entries):
        result = []

        if not isinstance(entries, list):
            return result

        for entry in entries:
            if "plugin" not in entry:
                continue

            if "octoversions" in entry and not is_octoprint_compatible(
                *entry["octoversions"]
            ):
                continue

            if "version" in entry:
                logger.debug(
                    "Blacklisted plugin: {}, version: {}".format(
                        entry["plugin"], entry["version"]
                    )
                )
                result.append((entry["plugin"], entry["version"]))
            elif "versions" in entry:
                logger.debug(
                    "Blacklisted plugin: {}, versions: {}".format(
                        entry["plugin"], ", ".join(entry["versions"])
                    )
                )
                for version in entry["versions"]:
                    result.append((entry["plugin"], version))
            else:
                logger.debug("Blacklisted plugin: {}".format(entry["plugin"]))
                result.append(entry["plugin"])

        return result

    def fetch_blacklist_from_cache(path, ttl):
        if not os.path.isfile(path):
            return None

        if os.stat(path).st_mtime + ttl < time.time():
            return None

        with bom_aware_open(path, encoding="utf-8", mode="rt") as f:
            result = yaml.safe_load(f)

        if isinstance(result, list):
            return result

    def fetch_blacklist_from_url(url, timeout=3, cache=None):
        result = []
        try:
            r = requests.get(url, timeout=timeout)
            result = process_blacklist(r.json())

            if cache is not None:
                try:
                    with bom_aware_open(cache, encoding="utf-8", mode="wt") as f:
                        yaml.safe_dump(result, f)
                except Exception as e:
                    logger.info(
                        "Fetched plugin blacklist but couldn't write it to its cache file: %s",
                        e,
                    )
        except Exception as e:
            logger.info(
                "Unable to fetch plugin blacklist from {}, proceeding without it: {}".format(
                    url, e
                )
            )
        return result

    try:
        # first attempt to fetch from cache
        cache_path = os.path.join(settings.getBaseFolder("data"), "plugin_blacklist.yaml")
        ttl = settings.getInt(["server", "pluginBlacklist", "ttl"])
        blacklist = fetch_blacklist_from_cache(cache_path, ttl)

        if blacklist is None:
            # no result from the cache, let's fetch it fresh
            url = settings.get(["server", "pluginBlacklist", "url"])
            timeout = settings.getFloat(["server", "pluginBlacklist", "timeout"])
            blacklist = fetch_blacklist_from_url(url, timeout=timeout, cache=cache_path)

        if blacklist is None:
            # still now result, so no blacklist
            blacklist = []

        if blacklist:
            logger.info(
                "Blacklist processing done, "
                "adding {} blacklisted plugin versions: {}".format(
                    len(blacklist), format_blacklist(blacklist)
                )
            )
        else:
            logger.info("Blacklist processing done")

        return blacklist
    except Exception:
        logger.exception(
            "Something went wrong while processing the plugin blacklist. Proceeding without it."
        )


def init_event_manager(settings):
    from octoprint.events import eventManager

    return eventManager()


def init_connectivity_checker(settings, event_manager):
    from octoprint.events import Events
    from octoprint.util import ConnectivityChecker

    # start regular check if we are connected to the internet
    connectivityEnabled = settings.getBoolean(["server", "onlineCheck", "enabled"])
    connectivityInterval = settings.getInt(["server", "onlineCheck", "interval"])
    connectivityHost = settings.get(["server", "onlineCheck", "host"])
    connectivityPort = settings.getInt(["server", "onlineCheck", "port"])
    connectivityName = settings.get(["server", "onlineCheck", "name"])

    def on_connectivity_change(
        old_value, new_value, connection_working=None, resolution_working=None
    ):
        event_manager.fire(
            Events.CONNECTIVITY_CHANGED,
            payload={
                "old": old_value,
                "new": new_value,
                "connection": connection_working,
                "resolution": resolution_working,
            },
        )

    connectivityChecker = ConnectivityChecker(
        connectivityInterval,
        connectivityHost,
        port=connectivityPort,
        name=connectivityName,
        enabled=connectivityEnabled,
        on_change=on_connectivity_change,
    )
    connectivityChecker.check_immediately()
    connectivityChecker.log_full_report()

    return connectivityChecker


def init_environment_detector(plugin_manager):
    from octoprint.environment import EnvironmentDetector

    return EnvironmentDetector(plugin_manager)


# ~~ server main method


def main():
    import sys

    # os args are gained differently on win32
    try:
        from click.utils import get_os_args

        args = get_os_args()
    except ImportError:
        # for whatever reason we are running an older Click version?
        args = sys.argv[1:]

    if len(args) >= len(sys.argv):
        # Now some ugly preprocessing of our arguments starts. We have a somewhat difficult situation on our hands
        # here if we are running under Windows and want to be able to handle utf-8 command line parameters (think
        # plugin parameters such as names or something, e.g. for the "dev plugin:new" command) while at the same
        # time also supporting sys.argv rewriting for debuggers etc (e.g. PyCharm).
        #
        # So what we try to do here is solve this... Generally speaking, sys.argv and whatever Windows returns
        # for its CommandLineToArgvW win32 function should have the same length. If it doesn't however and
        # sys.argv is shorter than the win32 specific command line arguments, obviously stuff was cut off from
        # sys.argv which also needs to be cut off of the win32 command line arguments.
        #
        # So this is what we do here.

        # -1 because first entry is the script that was called
        sys_args_length = len(sys.argv) - 1

        # cut off stuff from the beginning
        args = args[-1 * sys_args_length :] if sys_args_length else []

    from octoprint.util.fixes import patch_sarge_async_on_py2

    patch_sarge_async_on_py2()

    from octoprint.cli import octo

    octo(args=args, prog_name="octoprint", auto_envvar_prefix="OCTOPRINT")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2019 The OctoPrint Project - Released under terms of the AGPLv3 License"

import logging

import pkg_resources
import requests

from octoprint.util import to_native_str  # noqa: F401
from octoprint.util.version import (
    get_comparable_version,
    is_prerelease,
    is_python_compatible,
)

INFO_URL = "https://pypi.org/pypi/{package}/json"

logger = logging.getLogger("octoprint.plugins.softwareupdate.version_checks.pypi_release")


def _filter_out_latest(releases, include_prerelease=False, python_version=None):
    """
    Filters out the newest of all matching releases.

    Tests:

        >>> requires_py2 = ">=2.7.9,<3"
        >>> requires_py23 = ">=2.7.9, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, <4"
        >>> requires_py3 = ">=3.6, <4"
        >>> releases = {"1.3.12": [dict(requires_python=requires_py2, upload_time_iso_8601="2019-10-22T10:06:03.190293Z")], "1.4.0rc1": [dict(requires_python=requires_py23, upload_time_iso_8601="2019-11-22T10:06:03.190293Z")], "2.0.0rc1": [dict(requires_python=requires_py3, upload_time_iso_8601="2020-10-22T10:06:03.190293Z")]}
        >>> to_native_str(_filter_out_latest(releases, python_version="2.7.9"))
        '1.3.12'
        >>> to_native_str(_filter_out_latest(releases, include_prerelease=True, python_version="2.7.9"))
        '1.4.0rc1'
        >>> to_native_str(_filter_out_latest(releases, include_prerelease=True, python_version="3.6.0"))
        '2.0.0rc1'
        >>> to_native_str(_filter_out_latest(releases, python_version="3.6.0"))
    """
    releases = [{"version": k, "data": v[0]} for k, v in releases.items()]

    # filter out prereleases and versions incompatible to our python
    filter_function = lambda release: not is_prerelease(
        release["version"]
    ) and is_python_compatible(
        release["data"].get("requires_python", ""), python_version=python_version
    )
    if include_prerelease:
        filter_function = lambda release: is_python_compatible(
            release["data"].get("requires_python", ""), python_version=python_version
        )

    releases = list(filter(filter_function, releases))
    if not releases:
        return None

    # sort by upload date
    releases = sorted(
        releases, key=lambda release: release["data"].get("upload_time_iso_8601", "")
    )

    # latest release = last in list
    latest = releases[-1]

    return latest["version"]


def _get_latest_release(package, include_prerelease):
    from ..exceptions import NetworkError

    try:
        r = requests.get(INFO_URL.format(package=package), timeout=(3.05, 30))
    except requests.ConnectionError as exc:
        raise NetworkError(cause=exc)

    if not r.status_code == requests.codes.ok:
        return None

    data = r.json()
    if "info" not in data or "version" not in data["info"]:
        return None

    requires_python = data["info"].get("requires_python")
    if requires_python and not is_python_compatible(requires_python):
        return None

    return _filter_out_latest(data["releases"], include_prerelease=include_prerelease)


def _is_current(release_information):
    if release_information["remote"]["value"] is None:
        return True

    local_version = get_comparable_version(release_information["local"]["value"])
    remote_version = get_comparable_version(release_information["remote"]["value"])

    return remote_version <= local_version


def get_latest(target, check, online=True, *args, **kwargs):
    from ..exceptions import CannotUpdateOffline

    if not online and not check.get("offline", False):
        raise CannotUpdateOffline()

    package = check.get("package")

    distribution = pkg_resources.get_distribution(package)
    if distribution:
        local_version = distribution.version
    else:
        local_version = None

    remote_version = _get_latest_release(
        package, include_prerelease=check.get("prerelease", False)
    )

    information = {
        "local": {"name": local_version, "value": local_version},
        "remote": {"name": remote_version, "value": remote_version},
    }

    logger.debug(
        "Target: {}, local: {}, remote: {}".format(
            target, information["local"]["name"], information["remote"]["name"]
        )
    )

    return information, _is_current(information)
